"""Tests for alignment module.
"""

import random
import unittest
from math import log2

from adapt import alignment
from adapt.utils import lsh

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestAlignment(unittest.TestCase):
    """Tests methods in the Alignment class.
    """

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.a_seqs = ['ATCGAA', 'ATCGAT', 'AYCGAA', 'AYCGAT', 'AGCGAA']
        self.a = alignment.Alignment.from_list_of_seqs(self.a_seqs)

        self.b_seqs = ['ATCGAA', 'ATNNAT', 'ATCGNN', 'ATNNAT', 'ATNNAC']
        self.b = alignment.Alignment.from_list_of_seqs(self.b_seqs)

        self.c_seqs = ['ATCGAA', 'ATC-AA']
        self.c = alignment.Alignment.from_list_of_seqs(self.c_seqs)

        self.d_seqs = ['ATCGAA',
                       'ATCGAA',
                       'GGGCCC',
                       'ATCGAA',
                       'ATCGAA',
                       'ATCGAA',
                       'GGGCCC']
        self.d_seqs_aln = alignment.Alignment.from_list_of_seqs(self.d_seqs)

        self.gc = alignment.SequenceClusterer(lsh.HammingDistanceFamily(4), k=2)

    def test_make_list_of_seqs_a(self):
        self.assertEqual(self.a.make_list_of_seqs(),
                         self.a_seqs)
        self.assertEqual(self.a.make_list_of_seqs([0]),
                         ['ATCGAA'])
        self.assertEqual(self.a.make_list_of_seqs([0,3]),
                         ['ATCGAA', 'AYCGAT'])

    def test_make_list_of_seqs_b(self):
        self.assertEqual(self.b.make_list_of_seqs(),
                         self.b_seqs)

    def test_make_list_of_seqs_c(self):
        self.assertEqual(self.c.make_list_of_seqs(),
                         self.c_seqs)

    def test_make_list_of_seqs_c_with_no_gaps(self):
        self.assertEqual(self.c.make_list_of_seqs(remove_gaps=True),
                         ['ATCGAA', 'ATCAA'])

    def test_determine_consensus_sequence_a(self):
        self.assertEqual(self.a.determine_consensus_sequence(), 'ATCGAA')
        self.assertEqual(self.a.determine_consensus_sequence([0]), 'ATCGAA')
        self.assertIn(self.a.determine_consensus_sequence([2]),
                      ['ATCGAA', 'ACCGAA'])
        self.assertIn(self.a.determine_consensus_sequence([0,1]),
                      ['ATCGAA', 'ATCGAT'])
        self.assertIn(self.a.determine_consensus_sequence([1,2]),
                      ['ATCGAT', 'ATCGAA'])

    def test_determine_consensus_sequence_b(self):
        self.assertEqual(self.b.determine_consensus_sequence(), 'ATCGAT')

    def test_determine_consensus_sequence_c(self):
        with self.assertRaises(ValueError):
            # Should fail when determining consensus sequence given an indel
            self.c.determine_consensus_sequence()
        self.assertIn(self.c.determine_consensus_sequence([0]), 'ATCGAA')

    def test_seqs_with_gap(self):
        self.assertCountEqual(self.a.seqs_with_gap(), [])
        self.assertCountEqual(self.b.seqs_with_gap(), [])
        self.assertCountEqual(self.c.seqs_with_gap(), [1])

    def test_seqs_with_required_flanking_none_required(self):
        # No required flanking sequences should yield that all match
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                2, 3, (None, None)),
            {0,1,2,3,4,5,6}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                0, 3, (None, None)),
            {0,1,2,3,4,5,6}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                3, 3, (None, None)),
            {0,1,2,3,4,5,6}
        )

    def test_seqs_with_required_flanking_subset_to_consider(self):
        # Only look over a subset of the input sequences
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                2, 3, (None, None), seqs_to_consider={2,3,6}),
            {2,3,6}
        )

    def test_seqs_with_required_flanking_end5(self):
        # Required flanking on 5' end
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                3, 3, ('ATC', None)),
            {0,1,3,4,5}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                3, 3, ('C', None)),
            {0,1,3,4,5}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                3, 3, ('C', None), seqs_to_consider={0,1,3,4,5}),
            {0,1,3,4,5}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                3, 3, ('A', None)),
            {}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                3, 3, ('S', None)),
            {0,1,2,3,4,5,6}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                3, 3, ('NC', None)),
            {0,1,3,4,5}
        )

    def test_seqs_with_required_flanking_end5_too_close(self):
        # Required flanking on 5' end, with guide too close to end
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                1, 3, ('ATC', None)),
            {}
        )

    def test_seqs_with_required_flanking_end3(self):
        # Required flanking on 3' end
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                0, 3, (None, 'CCC')),
            {2,6}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                0, 3, (None, 'CNN')),
            {2,6}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                0, 3, (None, 'SNN')),
            {0,1,2,3,4,5,6}
        )

    def test_seqs_with_required_flanking_end3_too_close(self):
        # Required flanking on 3' end, with guide too close to end
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                1, 3, (None, 'CCC')),
            {}
        )

    def test_seqs_with_required_flanking_both_ends(self):
        # Required flanking on both ends
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                2, 2, ('G', 'AA')),
            {}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                2, 2, ('T', 'MN')),
            {0,1,3,4,5}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                2, 2, ('T', 'MN'), seqs_to_consider={1,2,3,4}),
            {1,3,4}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                2, 2, ('K', 'MN')),
            {0,1,2,3,4,5,6}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                2, 2, ('K', 'MN'), seqs_to_consider={4,5}),
            {4,5}
        )

    def test_seqs_with_required_flanking_both_ends_too_close(self):
        # Required flanking on both ends, with guide too close to end
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                3, 2, ('K', 'MN')),
            {}
        )
        self.assertCountEqual(self.d_seqs_aln.seqs_with_required_flanking(
                0, 2, ('K', 'MN')),
            {}
        )

    def test_construct_guide_a(self):
        self.assertEqual(self.a.construct_guide(0, 4, {0: {0,1,2,3,4}}, 0, False, self.gc),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a.construct_guide(0, 4, {0: {0,1,2,3,4}}, 1, False, self.gc),
                         ('ATCG', {0,1,2,3,4}))
        self.assertEqual(self.a.construct_guide(0, 4, {0: {4}}, 1, False, self.gc),
                         ('AGCG', {4}))
        self.assertIn(self.a.construct_guide(0, 4, {0: {2,3}}, 0, False, self.gc),
                      [('ATCG', {2,3}), ('ACCG', {2,3})])
        self.assertEqual(self.a.construct_guide(1, 4, {0: {0,1,2,3,4}}, 0, False, self.gc),
                         ('TCGA', {0,1,2,3}))
        self.assertEqual(self.a.construct_guide(2, 4, {0: {0,1,2,3,4}}, 0, False, self.gc),
                         ('CGAA', {0,2,4}))
        self.assertEqual(self.a.construct_guide(2, 4, {0: {0,1,2,3,4}}, 1, False, self.gc),
                         ('CGAA', {0,1,2,3,4}))
        self.assertEqual(self.a.construct_guide(2, 4, {0: {0,1,2,3,4}}, 2, False, self.gc),
                         ('CGAA', {0,1,2,3,4}))
        self.assertIn(self.a.construct_guide(2, 4, {0: {0,1,2,3}}, 0, False, self.gc),
                      [('CGAA', {0,2}), ('CGAT', {1,3})])
        self.assertIn(self.a.construct_guide(2, 4, {0: {0,1,2,3}}, 1, False, self.gc),
                      [('CGAA', {0,1,2,3}), ('CGAT', {0,1,2,3})])

    def test_construct_guide_b(self):
        # self.b has many Ns, which makes it difficult to write test cases
        # when clustering (the clusters tend to consist of guides in
        # which a position only has N); so pass None to guide_clusterer in
        # construct_guide() to skip clustering
        self.assertEqual(self.b.construct_guide(0, 4, {0: {0,1,2,3,4}}, 0, False, None),
                         ('ATCG', {0,2}))
        self.assertEqual(self.b.construct_guide(0, 4, {0: {0,1,2,3,4}}, 1, False, None),
                         ('ATCG', {0,2}))
        self.assertEqual(self.b.construct_guide(0, 4, {0: {0,1,2,3,4}}, 2, False, None),
                         ('ATCG', {0,1,2,3,4}))
        self.assertEqual(self.b.construct_guide(2, 4, {0: {0,1,2,3,4}}, 0, False, None),
                         ('CGAA', {0}))
        self.assertEqual(self.b.construct_guide(2, 4, {0: {0,1,2,3,4}}, 1, False, None),
                         ('CGAT', {0}))
        self.assertEqual(self.b.construct_guide(2, 4, {0: {0,1,2,3,4}}, 2, False, None),
                         ('CGAT', {0,1,2,3}))
        self.assertEqual(self.b.construct_guide(2, 4, {0: {0,1,2,3,4}}, 3, False, None),
                         ('CGAT', {0,1,2,3,4}))
        self.assertEqual(self.b.construct_guide(2, 4, {0: {2,4}}, 2, False, None),
                         ('CGAC', {2,4}))
        self.assertIn(self.b.construct_guide(2, 4, {0: {2,3,4}}, 2, False, None),
                      [('CGAC', {2,4}), ('CGAT', {2,3})])
        with self.assertRaises(alignment.CannotConstructGuideError):
            # Should fail when 'N' is all that exists at a position
            self.b.construct_guide(0, 4, {0: {1,3,4}}, 0, False, None)
        with self.assertRaises(alignment.CannotConstructGuideError):
            # Should fail when a potential guide (here, 'CGAC') cannot
            # bind to any sequence because they all have 'N' somewhere
            self.b.construct_guide(2, 4, {0: {2,4}}, 1, False, None)

    def test_construct_guide_c(self):
        with self.assertRaises(alignment.CannotConstructGuideError):
            # Should fail when the only sequence given (1) has an indel
            self.c.construct_guide(0, 4, {0: {1}}, 0, False, self.gc)

    def test_construct_guide_with_large_group_needed(self):
        seqs = ['ATCGAA',
                'ATCGAA',
                'GGGCCC',
                'ATCGAA',
                'ATCGAA',
                'ATCGAA',
                'GGGCCC']
        seqs_aln = alignment.Alignment.from_list_of_seqs(seqs)

        seqs_to_consider = {0: {0, 1, 3, 4, 5}, 1: {2, 6}}
        num_needed = {0: 3, 1: 1}
        # 'ATCGAA' is most sequences, and let's construct a guide by
        # needing more from the group consisting of these sequences
        self.assertEqual(seqs_aln.construct_guide(0, 4, seqs_to_consider, 0,
                            False, self.gc, num_needed=num_needed),
                         ('ATCG', {0, 1, 3, 4, 5}))

    def test_construct_guide_with_small_group_needed(self):
        seqs = ['ATCGAA',
                'ATCGAA',
                'GGGCCC',
                'ATCGAA',
                'ATCGAA',
                'ATCGAA',
                'GGGCCC']
        seqs_aln = alignment.Alignment.from_list_of_seqs(seqs)

        seqs_to_consider = {0: {0, 1, 3, 4, 5}, 1: {2, 6}}
        num_needed = {0: 1, 1: 2}
        # 'ATCGAA' is most sequences, but let's construct a guide by
        # needing more from a group consisting of the 'GGGCCC' sequences
        self.assertEqual(seqs_aln.construct_guide(0, 4, seqs_to_consider, 0,
                            False, self.gc, num_needed=num_needed),
                         ('GGGC', {2, 6}))

    def test_construct_guide_with_suitable_fn(self):
        seqs = ['GTATCAAAT',
                'CTACCAAAA',
                'GTATCAAAT',
                'GTATCAAAT']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        guide_length = 6
        seqs_to_consider = {0: {0, 1, 2, 3}}
        guide_clusterer = alignment.SequenceClusterer(
            lsh.HammingDistanceFamily(guide_length),
            k=3)

        # The best guide is 'GTATCA'
        p = aln.construct_guide(0, guide_length, seqs_to_consider, 1, False, guide_clusterer)
        gd, covered_seqs = p
        self.assertEqual(gd, 'GTATCA')
        self.assertEqual(covered_seqs, {0, 2, 3})

        # Do not allow guides with 'TAT' in them
        def f(guide):
            if 'TAT' in guide:
                return False
            else:
                return True
        # Now the best guide is 'CTACCA'
        p = aln.construct_guide(0, guide_length, seqs_to_consider, 1, False, guide_clusterer,
            guide_is_suitable_fn=f)
        gd, covered_seqs = p
        self.assertEqual(gd, 'CTACCA')
        self.assertEqual(covered_seqs, {1})

        # Do not allow guides with 'A' in them
        def f(guide):
            if 'A' in guide:
                return False
            else:
                return True
        # Now there is no suitable guide
        with self.assertRaises(alignment.CannotConstructGuideError):
            aln.construct_guide(0, guide_length, seqs_to_consider, 1, False, guide_clusterer,
                guide_is_suitable_fn=f)

    def test_construct_guide_with_predictor(self):
        seqs = ['GTATCAAAT',
                'ATACCAAAA',
                'GTATCAAAT',
                'GTATCAAAT']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        guide_length = 6
        seqs_to_consider = {0: {0, 1, 2, 3}}
        guide_clusterer = alignment.SequenceClusterer(
            lsh.HammingDistanceFamily(guide_length),
            k=3)

        # The best guide is 'GTATCA'
        p = aln.construct_guide(0, guide_length, seqs_to_consider, 1, False, guide_clusterer)
        gd, covered_seqs = p
        self.assertEqual(gd, 'GTATCA')
        self.assertEqual(covered_seqs, {0, 2, 3})

        # Only predict guides starting with 'A' to be active
        class PredictorTest:
            def __init__(self):
                self.context_nt = 0
            def determine_highly_active(self, start_pos, pairs):
                y = []
                for target, guide in pairs:
                    y += [guide[0] == 'A']
                return y
        predictor = PredictorTest()
        # Now the best guide is 'ATACCA'
        p = aln.construct_guide(0, guide_length, seqs_to_consider, 1, False, guide_clusterer,
            predictor=predictor, stop_early=False)
        gd, covered_seqs = p
        self.assertEqual(gd, 'ATACCA')
        self.assertEqual(covered_seqs, {1})

        # Only predict guides starting with 'A' to be active, and impose an
        # early stopping criterion
        class PredictorTest:
            def __init__(self):
                self.context_nt = 0
            def determine_highly_active(self, start_pos, pairs):
                y = []
                for target, guide in pairs:
                    y += [guide[0] == 'A']
                return y
        predictor = PredictorTest()
        # With early stopping, it will not find a guide
        with self.assertRaises(alignment.CannotConstructGuideError):
            aln.construct_guide(0, guide_length, seqs_to_consider, 1, False, guide_clusterer,
                predictor=predictor, stop_early=True)

        # Only predictor guides starting with 'C' to be active
        class PredictorTest:
            def __init__(self):
                self.context_nt = 0
            def determine_highly_active(self, start_pos, pairs):
                y = []
                for target, guide in pairs:
                    y += [guide[0] == 'C']
                return y
        predictor = PredictorTest()
        # Now there is no suitable guide
        with self.assertRaises(alignment.CannotConstructGuideError):
            aln.construct_guide(0, guide_length, seqs_to_consider, 1, False, guide_clusterer,
                predictor=predictor)

    def test_construct_guide_with_required_flanking(self):
        seqs = ['TCAAAT',
                'CCAAAA',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'TCAAAT',
                'TCAAAT']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        guide_length = 2
        guide_clusterer = alignment.SequenceClusterer(
            lsh.HammingDistanceFamily(guide_length),
            k=2)
        seqs_to_consider = {0: set(range(len(seqs)))}

        # The best guide at start=2 is 'TT', but if we require
        # 'C' to flank on the 5' end, the best is 'AA'
        p = aln.construct_guide(2, guide_length, seqs_to_consider, 1, False,
            guide_clusterer, required_flanking_seqs=('C', None))
        gd, covered_seqs = p
        self.assertEqual(gd, 'AA')
        self.assertEqual(covered_seqs, {0,1,9,10})

        # The best guide at start=2 is 'TT', but if we require
        # 'C' to flank on the 5' end, the best is 'AA'
        # Now if we require 'M' on the 5' end, 'TT' will be the best guide
        p = aln.construct_guide(2, guide_length, seqs_to_consider, 1, False,
            guide_clusterer, required_flanking_seqs=('M', None))
        gd, covered_seqs = p
        self.assertEqual(gd, 'TT')
        self.assertEqual(covered_seqs, {2,3,4,5,6,7,8})

    def test_determine_representative_guides(self):
        seqs = ['TCAAAT',
                'CCAAAA',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'GGGGGG',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'TCAAAT',
                'TCAAAT',
                'TCAAAA',
                'T-GGAA']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        guide_length = 6
        guide_clusterer = alignment.SequenceClusterer(
            lsh.HammingDistanceFamily(guide_length),
            k=3)
        seqs_to_consider = {0: set(range(len(seqs)))}

        representatives = aln.determine_representative_guides(0,
                guide_length, seqs_to_consider, guide_clusterer)
        self.assertSetEqual(representatives,
                {'TCAAAT', 'CCAAAA', 'GGGGGG', 'CATTTT'})

    def test_compute_activity(self):
        # Predict guides matching target to have activity 1, and
        # starting with 'A' to have activity 2 (otherwise, 0)
        class PredictorTest:
            def __init__(self):
                self.context_nt = 1
            def compute_activity(self, start_pos, pairs):
                y = []
                for target, guide in pairs:
                    target_without_context = target[self.context_nt:len(target)-self.context_nt]
                    if guide == target_without_context:
                        if guide[0] == 'A':
                            y += [2]
                        else:
                            y += [1]
                    else:
                        y += [0]
                return y
        predictor = PredictorTest()

        seqs = ['TCAAAT',
                'CCAAAA',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'GGGGGG',
                'CATTTT',
                'CATTTT',
                'CAT-TT',
                'TCAAAT',
                'TCA-AT',
                'TCAAAA',
                'TCGGAA']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        activities_AAA = aln.compute_activity(2, 'AAA', predictor)
        self.assertListEqual(list(activities_AAA),
                [2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0])

        activities_TTT = aln.compute_activity(2, 'TTT', predictor)
        self.assertListEqual(list(activities_TTT),
                [0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0])

    def test_sequences_bound_by_guide(self):
        seqs = ['TCAAAT',
                'CCAAAA',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CATTTT',
                'CCTTGT',
                'TCAAAT',
                'TCAAAT']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        self.assertEqual(aln.sequences_bound_by_guide('TT', 2, 0, False,
                            required_flanking_seqs=(None, None)),
                         [2,3,4,5,6,7,8,9])
        self.assertEqual(aln.sequences_bound_by_guide('TT', 2, 0, False,
                            required_flanking_seqs=('A', None)),
                         [2,3,4,5,6,7,8])
        self.assertEqual(aln.sequences_bound_by_guide('TT', 2, 0, False,
                            required_flanking_seqs=('C', None)),
                         [9])
        self.assertEqual(aln.sequences_bound_by_guide('TT', 2, 0, False,
                            required_flanking_seqs=(None, 'G')),
                         [9])
        self.assertEqual(aln.sequences_bound_by_guide('TT', 2, 0, False,
                            required_flanking_seqs=(None, 'GN')),
                         [9])
        self.assertEqual(aln.sequences_bound_by_guide('TT', 2, 0, False,
                            required_flanking_seqs=(None, 'N')),
                         [2,3,4,5,6,7,8,9])

    def test_sequences_bound_by_guide_with_required_flanking(self):
        self.assertEqual(self.a.sequences_bound_by_guide('ATCG', 0, 0, False),
                         [0,1,2,3])
        self.assertEqual(self.a.sequences_bound_by_guide('ATCG', 0, 1, False),
                         [0,1,2,3,4])

    def test_most_common_sequence_simple(self):
        self.assertEqual(self.d_seqs_aln.determine_most_common_sequence(
                            skip_ambiguity=False),
                         'ATCGAA')
        self.assertEqual(self.d_seqs_aln.determine_most_common_sequence(
                            skip_ambiguity=True),
                         'ATCGAA')

    def test_most_common_sequence_with_ambiguity(self):
        seqs = ['ATCNAA',
                'ATCNAA',
                'GGGCCC',
                'ATCGAA',
                'ATCNAA',
                'ATCNAA',
                'GGGCCC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        self.assertEqual(aln.determine_most_common_sequence(skip_ambiguity=False),
                         'ATCNAA')
        self.assertEqual(aln.determine_most_common_sequence(skip_ambiguity=True),
                         'GGGCCC')

    def test_position_entropy_simple(self):
        seqs = ['ACCCC',
                'AAGGC',
                'AAATA',
                'AAAAA']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        all_ps = [[1],
                  [0.25, 0.75],
                  [0.25, 0.25, .5],
                  [0.25, 0.25, 0.25, 0.25],
                  [0.5, 0.5]]

        entropy = [sum([-p*log2(p) for p in ps]) for ps in all_ps]
        self.assertEqual(aln.position_entropy(), entropy)

    def test_position_entropy_with_ambiguity(self):
        seqs = ['MRWSYKVHDBN-']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        all_ps = [[0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [1/3.0, 1/3.0, 1/3.0],
                  [1/3.0, 1/3.0, 1/3.0],
                  [1/3.0, 1/3.0, 1/3.0],
                  [1/3.0, 1/3.0, 1/3.0],
                  [0.25, 0.25, 0.25, 0.25],
                  [1]]

        entropy = [sum([-p*log2(p) for p in ps]) for ps in all_ps]
        self.assertEqual(aln.position_entropy(), entropy)

    def test_base_percentages_simple(self):
        seqs = ['ACCCC',
                'AAGGC',
                'AAATA',
                'AAAAA']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        base_p = {
            'A': 0.6,
            'C': 0.25,
            'G': 0.1,
            'T': 0.05
        }

        self.assertEqual(aln.base_percentages(), base_p)

    def test_base_percentages_with_ambiguity(self):
        seqs = ['MRWSYKVHDBN-']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        base_p = {
            'A': 0.25,
            'C': 0.25,
            'G': 0.25,
            'T': 0.25
        }

        self.assertEqual(aln.base_percentages(), base_p)

    def test_construct_from_0_seqs(self):
        with self.assertRaises(Exception):
            seqs = []
            alignment.Alignment.from_list_of_seqs(seqs)


class TestSequenceClusterer(unittest.TestCase):
    """Tests the SequenceClusterer class.
    """

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        family = lsh.HammingDistanceFamily(10)
        self.sc = alignment.SequenceClusterer(family, k=3)

    def test_cluster(self):
        seqs = [('ATCGAAATAA', 4),
                ('ATCGAAATAA', 5),
                ('ATTGAAATAT', 1),
                ('CTGGTCATAA', 2),
                ('CTCGTCATAA', 3)]

        clusters = self.sc.cluster(seqs)
        self.assertCountEqual(clusters,
            {frozenset({1, 4, 5}), frozenset({2, 3})})

    def test_largest_cluster(self):
        seqs = [('ATCGAAATAA', 4),
                ('ATCGAAATAA', 5),
                ('ATTGAAATAT', 1),
                ('CTGGTCATAA', 2),
                ('CTCGTCATAA', 3)]

        largest_cluster = self.sc.largest_cluster(seqs)
        self.assertCountEqual(largest_cluster, {1, 4, 5})




"""Tests for kmer_shard module.
"""

import unittest

from adapt.specificity import kmer_shard

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestTrieSpace(unittest.TestCase):

    def test_get(self):
        # Build and fetch from a trie space
        ts = kmer_shard.TrieSpace()
        t = ts.get('abc', make=True)
        t.insert([('A', 4)])
        t = ts.get('abc', make=False)
        self.assertEqual(t.query('A') , [4])
        self.assertIsNone(ts.get('def', make=False))


class TestSignatureFunctions(unittest.TestCase):

    def test_signatures_with_mismatches(self):
        self.assertCountEqual(kmer_shard._signatures_with_mismatches('GTG', 0),
                ['GTG'])
        self.assertCountEqual(kmer_shard._signatures_with_mismatches('GTG', 1),
                ['GTG', 'TTG', 'GGG', 'GTT'])
        self.assertCountEqual(kmer_shard._signatures_with_mismatches('GTG', 2),
                ['GTG', 'TTG', 'GGG', 'GTT', 'TGG', 'TTT', 'GGT'])

    def test_full_signature(self):
        self.assertEqual(kmer_shard._full_signature('ATCGATCG'),
                'GTTGGTTG')

    def test_full_signatures_with_mismatches(self):
        self.assertCountEqual(
                kmer_shard._full_signatures_with_mismatches('ATG', 1),
                ['GTG', 'TTG', 'GGG', 'GTT'])

    def test_split_signature(self):
        self.assertEqual(kmer_shard._split_signature('ATCG', 0),
                'GT')
        self.assertEqual(kmer_shard._split_signature('ATCG', 1),
                'TG')
        self.assertEqual(kmer_shard._split_signature('ATCGA', 0),
                'GTT')
        self.assertEqual(kmer_shard._split_signature('ATCGA', 1),
                'GG')

    def test_split_signatures_with_mismatches(self):
        self.assertCountEqual(
                kmer_shard._split_signatures_with_mismatches('ATCG', 0, 1),
                ['GT', 'TT', 'GG'])
        self.assertCountEqual(
                kmer_shard._split_signatures_with_mismatches('ATCG', 1, 1),
                ['TG', 'GG', 'TT'])


class TestTrieSpaceOfKmersEvenK(unittest.TestCase):
    """Test TrieSpaceOfKmersFullSig and TrieSpaceOfKmersSplitSig using
    k-mers with even length (i.e., k is even).

    This is useful to split up because the methods treat even and odd lengths
    slightly differently.
    """

    def build(self, cls):
        # cls is the class to instantiate
        tsok = cls()
        tsok.add([
            ('ATCGAT', {(0,1), (0,2), (0,4), (1,1)}),
            ('ATAAAT', {(0,3), (2,1)})
        ])
        tsok.add([
            ('GGGGGG', {(3,1), (3,2)}),
            ('TTCGAT', {(0,5), (1,2)})
        ])
        return tsok

    def query_tests(self, tsok):
        # Exact matches, no G-U pairing
        self.assertEqual(tsok.query('ATCGAT', m=0, gu_pairing=False),
                {0: {1,2,4}, 1: {1}})
        self.assertEqual(tsok.query('ATAAAT', m=0, gu_pairing=False),
                {0: {3}, 2: {1}})
        self.assertEqual(tsok.query('GGGGGG', m=0, gu_pairing=False),
                {3: {1,2}})
        self.assertEqual(tsok.query('TTCGAT', m=0, gu_pairing=False),
                {0: {5}, 1: {2}})
        self.assertEqual(tsok.query('CCCCCC', m=0, gu_pairing=False),
                {})

        # Queries 1 mismatch away, no G-U pairing
        self.assertEqual(tsok.query('ATCGAT', m=1, gu_pairing=False),
                {0: {1,2,4,5}, 1: {1,2}})
        self.assertEqual(tsok.query('GGTGGG', m=1, gu_pairing=False),
                {3: {1,2}})
        self.assertEqual(tsok.query('GGTTGG', m=1, gu_pairing=False),
                {})

        # Exact matches, with G-U pairing
        self.assertEqual(tsok.query('ACCGAT', m=0, gu_pairing=True),
                {0: {1,2,4}, 1: {1}})
        self.assertEqual(tsok.query('ACCAAT', m=0, gu_pairing=True),
                {0: {1,2,4}, 1: {1}})
        self.assertEqual(tsok.query('AAAAAA', m=0, gu_pairing=True),
                {3: {1,2}})

        # Queries 1 mismatch away, with G-U pairing
        self.assertEqual(tsok.query('ACCGAT', m=1, gu_pairing=True),
                {0: {1,2,4,5}, 1: {1,2}})
        self.assertEqual(tsok.query('GGTGGA', m=1, gu_pairing=True),
                {3: {1,2}})
        self.assertEqual(tsok.query('GGTTGG', m=1, gu_pairing=True),
                {})
        self.assertEqual(tsok.query('ACATAT', m=1, gu_pairing=True),
                {0: {3}, 2: {1}})
        self.assertEqual(tsok.query('AAATAT', m=1, gu_pairing=True),
                {})

        # 6 mismatches away should yield all results
        self.assertEqual(tsok.query('AAAAAA', m=6, gu_pairing=False),
                {0: {1,2,3,4,5}, 1: {1,2}, 2: {1}, 3: {1,2}})
        self.assertEqual(tsok.query('AAAAAA', m=6, gu_pairing=True),
                {0: {1,2,3,4,5}, 1: {1,2}, 2: {1}, 3: {1,2}})

    def test_with_full_signatures(self):
        tsok = self.build(kmer_shard.TrieSpaceOfKmersFullSig)
        self.query_tests(tsok)

    def test_with_split_signatures(self):
        tsok = self.build(kmer_shard.TrieSpaceOfKmersSplitSig)
        self.query_tests(tsok)


class TestTrieSpaceOfKmersOddK(unittest.TestCase):
    """Test TrieSpaceOfKmersFullSig and TrieSpaceOfKmersSplitSig using
    k-mers with odd length (i.e., k is odd).

    This is useful to split up because the methods treat even and odd lengths
    slightly differently.
    """

    # TODO make these k-mers be odd length

    def build(self, cls):
        # cls is the class to instantiate
        tsok = cls()
        tsok.add([
            ('ATCGAT', {(0,1), (0,2), (0,4), (1,1)}),
            ('ATAAAT', {(0,3), (2,1)})
        ])
        tsok.add([
            ('GGGGGG', {(3,1), (3,2)}),
            ('TTCGAT', {(0,5), (1,2)})
        ])
        return tsok

    def query_tests(self, tsok):
        # Exact matches, no G-U pairing
        self.assertEqual(tsok.query('ATCGAT', m=0, gu_pairing=False),
                {0: {1,2,4}, 1: {1}})
        self.assertEqual(tsok.query('ATAAAT', m=0, gu_pairing=False),
                {0: {3}, 2: {1}})
        self.assertEqual(tsok.query('GGGGGG', m=0, gu_pairing=False),
                {3: {1,2}})
        self.assertEqual(tsok.query('TTCGAT', m=0, gu_pairing=False),
                {0: {5}, 1: {2}})
        self.assertEqual(tsok.query('CCCCCC', m=0, gu_pairing=False),
                {})

        # Queries 1 mismatch away, no G-U pairing
        self.assertEqual(tsok.query('ATCGAT', m=1, gu_pairing=False),
                {0: {1,2,4,5}, 1: {1,2}})
        self.assertEqual(tsok.query('GGTGGG', m=1, gu_pairing=False),
                {3: {1,2}})
        self.assertEqual(tsok.query('GGTTGG', m=1, gu_pairing=False),
                {})

        # Exact matches, with G-U pairing
        self.assertEqual(tsok.query('ACCGAT', m=0, gu_pairing=True),
                {0: {1,2,4}, 1: {1}})
        self.assertEqual(tsok.query('ACCAAT', m=0, gu_pairing=True),
                {0: {1,2,4}, 1: {1}})
        self.assertEqual(tsok.query('AAAAAA', m=0, gu_pairing=True),
                {3: {1,2}})

        # Queries 1 mismatch away, with G-U pairing
        self.assertEqual(tsok.query('ACCGAT', m=1, gu_pairing=True),
                {0: {1,2,4,5}, 1: {1,2}})
        self.assertEqual(tsok.query('GGTGGA', m=1, gu_pairing=True),
                {3: {1,2}})
        self.assertEqual(tsok.query('GGTTGG', m=1, gu_pairing=True),
                {})
        self.assertEqual(tsok.query('ACATAT', m=1, gu_pairing=True),
                {0: {3}, 2: {1}})
        self.assertEqual(tsok.query('AAATAT', m=1, gu_pairing=True),
                {})

        # 6 mismatches away should yield all results
        self.assertEqual(tsok.query('AAAAAA', m=6, gu_pairing=False),
                {0: {1,2,3,4,5}, 1: {1,2}, 2: {1}, 3: {1,2}})
        self.assertEqual(tsok.query('AAAAAA', m=6, gu_pairing=True),
                {0: {1,2,3,4,5}, 1: {1,2}, 2: {1}, 3: {1,2}})

    def test_with_full_signatures(self):
        tsok = self.build(kmer_shard.TrieSpaceOfKmersFullSig)
        self.query_tests(tsok)

    def test_with_split_signatures(self):
        tsok = self.build(kmer_shard.TrieSpaceOfKmersSplitSig)
        self.query_tests(tsok)



"""Tests for guide module.
"""

import unittest

from adapt.utils import guide

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestSeqMismatches(unittest.TestCase):
    """Tests the seq_mismatches function.
    """

    def test_equal(self):
        self.assertEqual(guide.seq_mismatches('ATCG', 'ATCG'), 0)

    def test_equal_with_ambiguity(self):
        self.assertEqual(guide.seq_mismatches('AYCG', 'ATCG'), 0)
        self.assertEqual(guide.seq_mismatches('AYCG', 'AYCG'), 0)
        self.assertEqual(guide.seq_mismatches('AYCG', 'AYCB'), 0)
        self.assertEqual(guide.seq_mismatches('AYCG', 'AYSB'), 0)

    def test_one_mismatch(self):
        self.assertEqual(guide.seq_mismatches('ATCG', 'AACG'), 1)
        self.assertEqual(guide.seq_mismatches('ATCG', 'ATCA'), 1)
        self.assertEqual(guide.seq_mismatches('ATCG', 'ATCN'), 1)
        self.assertEqual(guide.seq_mismatches('ATCN', 'ATCN'), 1)
        self.assertEqual(guide.seq_mismatches('ATCN', 'ATCG'), 1)

    def test_one_mismatch_with_ambiguity(self):
        self.assertEqual(guide.seq_mismatches('AYCG', 'ATCA'), 1)
        self.assertEqual(guide.seq_mismatches('AYCG', 'ARCG'), 1)
        self.assertEqual(guide.seq_mismatches('AYCG', 'ADCA'), 1)
        self.assertEqual(guide.seq_mismatches('AYCG', 'ATCN'), 1)
        self.assertEqual(guide.seq_mismatches('AYCN', 'ATCG'), 1)

    def test_multiple_mismatches(self):
        self.assertEqual(guide.seq_mismatches('ATCG', 'AAAA'), 3)
        self.assertEqual(guide.seq_mismatches('ATCG', 'ANNN'), 3)

    def test_multiple_mismatches_with_ambiguity(self):
        self.assertEqual(guide.seq_mismatches('AYCG', 'ARCA'), 2)
        self.assertEqual(guide.seq_mismatches('AYCG', 'ATAA'), 2)
        self.assertEqual(guide.seq_mismatches('AYCG', 'ATNN'), 2)
        self.assertEqual(guide.seq_mismatches('AYNN', 'ATCG'), 2)


class TestSeqMismatchesWithGUPairing(unittest.TestCase):
    """Tests the seq_mismatches_with_gu_pairs function.
    """

    def test_equal(self):
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'ATCG'), 0)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATAG', 'ATGG'), 0)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'ATTG'), 0)

    def test_equal_with_ambiguity(self):
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATYG', 'ATCG'), 0)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATYG', 'ATBG'), 0)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATRG', 'ATGG'), 0)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATRG', 'ATSG'), 0)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'ATDG'), 0)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATMG', 'ATDG'), 0)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATSG', 'ATWG'), 0)

    def test_one_mismatch(self):
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'AACG'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'ATCA'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'ATCN'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCN', 'ATCN'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCN', 'ATCG'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATAN', 'ATAG'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('CTAC', 'TTAG'), 1)

    def test_one_mismatch_with_ambiguity(self):
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCG', 'ATCA'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCG', 'ARCG'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCG', 'ADCA'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCG', 'ATCN'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCN', 'ATCG'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCA', 'ATCT'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYAC', 'AGGT'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYAC', 'ARGT'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYAY', 'ARGT'), 1)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYAY', 'AYNT'), 1)

    def test_multiple_mismatches(self):
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'AAAA'), 3)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'ANNN'), 3)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'GNNN'), 3)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ATCG', 'GATA'), 2)

    def test_multiple_mismatches_with_ambiguity(self):
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCG', 'ARCA'), 2)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCG', 'ATAA'), 2)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYCG', 'ATNN'), 2)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYNN', 'ATCG'), 2)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYNN', 'ATCG'), 2)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('AYTG', 'ATCA'), 2)
        self.assertEqual(guide.seq_mismatches_with_gu_pairs('ACTG', 'ATCA'), 2)


class TestQueryTargetEq(unittest.TestCase):
    """Tests the query_target_eq function.
    """

    def test_is_equal(self):
        self.assertTrue(guide.query_target_eq('N', 'A'))
        self.assertTrue(guide.query_target_eq('H', 'A'))
        self.assertTrue(guide.query_target_eq('A', 'A'))
        self.assertTrue(guide.query_target_eq('ANN', 'AAA'))
        self.assertTrue(guide.query_target_eq('HNN', 'AAA'))
        self.assertTrue(guide.query_target_eq('HNA', 'TAA'))
        self.assertTrue(guide.query_target_eq('HNA', 'TAR'))
        self.assertTrue(guide.query_target_eq('HNT', 'TAH'))
        self.assertTrue(guide.query_target_eq('HNT', 'YAH'))

    def test_is_not_equal(self):
        self.assertFalse(guide.query_target_eq('N', 'N'))
        self.assertFalse(guide.query_target_eq('A', 'N'))
        self.assertFalse(guide.query_target_eq('A', 'T'))
        self.assertFalse(guide.query_target_eq('GAA', 'GAT'))
        self.assertFalse(guide.query_target_eq('HNN', 'GAA'))
        self.assertFalse(guide.query_target_eq('HNN', 'GAA'))
        self.assertFalse(guide.query_target_eq('HNN', 'AAN'))
        self.assertFalse(guide.query_target_eq('AAA', 'NNN'))
        self.assertFalse(guide.query_target_eq('AAA', 'YAA'))


class TestGuideBindsWithoutGUPairing(unittest.TestCase):
    """Tests the guide_binds function without allowing G-U pairs.
    """

    def test_guide_does_bind(self):
        self.assertTrue(guide.guide_binds('ATCG', 'ATCG', 0, False))
        self.assertTrue(guide.guide_binds('ATCG', 'AYCG', 0, False))
        self.assertTrue(guide.guide_binds('ATCG', 'AYCB', 0, False))
        self.assertTrue(guide.guide_binds('ATCG', 'ATCA', 1, False))
        self.assertTrue(guide.guide_binds('ATCG', 'ARCS', 1, False))
        self.assertTrue(guide.guide_binds('ATCG', 'AYCA', 1, False))
        self.assertTrue(guide.guide_binds('ATCG', 'AYCN', 1, False))
        self.assertTrue(guide.guide_binds('ATCG', 'AYNN', 2, False))

    def test_guide_does_not_bind(self):
        self.assertFalse(guide.guide_binds('ATCG', 'AACG', 0, False))
        self.assertFalse(guide.guide_binds('ATCG', 'ASCG', 0, False))
        self.assertFalse(guide.guide_binds('ATCG', 'AYNG', 0, False))
        self.assertFalse(guide.guide_binds('ATCG', 'ATTG', 0, False))
        self.assertFalse(guide.guide_binds('ATCG', 'ATAA', 1, False))
        self.assertFalse(guide.guide_binds('ATCG', 'ATNN', 1, False))


class TestGuideBindsWithGUPairing(unittest.TestCase):
    """Tests the guide_binds function while allowing G-U pairs.
    """

    def test_guide_does_bind(self):
        self.assertTrue(guide.guide_binds('ATCG', 'ATCG', 0, True))
        self.assertTrue(guide.guide_binds('ATCG', 'AYCG', 0, True))
        self.assertTrue(guide.guide_binds('ATCG', 'AYCB', 0, True))
        self.assertTrue(guide.guide_binds('ATCG', 'ATCA', 1, True))
        self.assertTrue(guide.guide_binds('ATCG', 'ARCS', 1, True))
        self.assertTrue(guide.guide_binds('ATCG', 'AYCA', 1, True))
        self.assertTrue(guide.guide_binds('ATCG', 'AYCN', 1, True))
        self.assertTrue(guide.guide_binds('ATCG', 'AYNN', 2, True))
        self.assertTrue(guide.guide_binds('AACG', 'AGCG', 0, True))
        self.assertTrue(guide.guide_binds('AACG', 'AGCT', 1, True))
        self.assertTrue(guide.guide_binds('ACGG', 'ATGG', 0, True))
        self.assertTrue(guide.guide_binds('ACGG', 'ATGT', 1, True))
        self.assertTrue(guide.guide_binds('AACC', 'GGTT', 0, True))
        self.assertTrue(guide.guide_binds('ATCG', 'GCTA', 2, True))
        self.assertTrue(guide.guide_binds('MTCG', 'RCTA', 2, True))

    def test_guide_does_not_bind(self):
        self.assertFalse(guide.guide_binds('ATCG', 'AACG', 0, True))
        self.assertFalse(guide.guide_binds('ATCG', 'ASCG', 0, True))
        self.assertFalse(guide.guide_binds('ATCG', 'AYNG', 0, True))
        self.assertFalse(guide.guide_binds('WTCG', 'CTCG', 0, True))
        self.assertFalse(guide.guide_binds('ATCG', 'ATAA', 1, True))
        self.assertFalse(guide.guide_binds('ATCG', 'ATNN', 1, True))
        self.assertFalse(guide.guide_binds('ATCG', 'GCTA', 1, True))
        self.assertFalse(guide.guide_binds('CTCG', 'GCTA', 2, True))
        self.assertFalse(guide.guide_binds('WTCG', 'CTCG', 0, True))


class TestGCFrac(unittest.TestCase):
    """Tests the gc_frac function.
    """

    def test_gc_frac(self):
        self.assertEqual(guide.gc_frac('ATTA'), 0)
        self.assertEqual(guide.gc_frac('GGCC'), 1)
        self.assertEqual(guide.gc_frac('AGTC'), 0.5)


class TestGuideOverlapInSeq(unittest.TestCase):
    """Tests the guide_overlap_in_seq() function.
    """

    def test_disjoint(self):
        self.assertEqual(guide.guide_overlap_in_seq(
            ['ATC'], 'AATTCCATCGG', 0, False),
            {6,7,8})
        self.assertEqual(guide.guide_overlap_in_seq(
            ['ATC', 'ATT'], 'AATTCCATCGG', 0, False),
            {1,2,3,6,7,8})

    def test_overlapping(self):
        self.assertEqual(guide.guide_overlap_in_seq(
            ['ATT', 'TCC', 'TCG'], 'AATTCCATCGG', 0, False),
            {1,2,3,4,5,7,8,9})

    def test_at_ends(self):
        self.assertEqual(guide.guide_overlap_in_seq(
            ['AAT', 'CGG'], 'AATTCCATCGG', 0, False),
            {0,1,2,8,9,10})

    def test_with_mismatches(self):
        self.assertEqual(guide.guide_overlap_in_seq(
            ['ATA'], 'AATTCCATCGG', 1, False),
            {1,2,3,6,7,8})



"""Tests for kmer_leaf module.
"""

import unittest

from adapt.specificity import kmer_leaf

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestKmerLeaf(unittest.TestCase):
    """Tests KmerLeaf objects.
    """
    
    def test_create(self):
        kl = kmer_leaf.KmerLeaf([(0,1),(0,2),(0,2),(3,5)])
        kl2 = kmer_leaf.KmerLeaf([(0,1),(0,2),(3,5)])
        self.assertEqual(len(kl), 3)
        self.assertTrue(kl == kl2)

    def test_extend(self):
        kl = kmer_leaf.KmerLeaf([(0,1),(0,2),(0,2),(3,5)])
        kl2 = kmer_leaf.KmerLeaf([(0,1),(3,6),(4,7)])
        self.assertEqual(len(kl), 3)
        kl.extend(kl2)
        self.assertEqual(len(kl), 5)
        self.assertEqual(len(kl2), 3)

    def test_contains(self):
        kl = kmer_leaf.KmerLeaf([(0,1),(0,2),(0,2),(3,5)])
        self.assertTrue(0 in kl)
        self.assertTrue(3 in kl)
        self.assertFalse(1 in kl)
    
    def test_union(self):
        kl = kmer_leaf.KmerLeaf([(0,1),(0,2),(0,2),(3,5)])
        kl2 = kmer_leaf.KmerLeaf([(0,1),(3,6),(4,7)])
        self.assertEqual(len(kl), 3)
        self.assertEqual(len(kl2), 3)
        kl_union = kmer_leaf.KmerLeaf.union([kl, kl2])
        self.assertEqual(len(kl), 3)
        self.assertEqual(len(kl2), 3)
        self.assertEqual(len(kl_union), 5)

    def test_remove(self):
        kl = kmer_leaf.KmerLeaf([(0,1),(0,2),(0,2),(3,5)])
        self.assertEqual(len(kl), 3)
        kl.remove(0)
        self.assertEqual(len(kl), 1)


    def test_copy(self):
        kl = kmer_leaf.KmerLeaf([(0,1),(0,2),(0,2),(3,5)])
        kl_copy = kl.copy()
        self.assertTrue(kl == kl_copy)
        kl.remove(0)
        self.assertEqual(len(kl), 1)
        self.assertEqual(len(kl_copy), 3)




"""Tests for formatting module.
"""

import logging
import unittest

from adapt.utils import formatting

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestTagSeqOverlap(unittest.TestCase):
    """Tests the tag_seq_overlap() function.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.WARNING)

    def t(self, overlap_labels, expect, target_seq='AATTCCATCGG'):
        f = formatting.tag_seq_overlap(overlap_labels, target_seq)
        self.assertEqual(f, expect)

    def test_disjoint(self):
        self.t({'pr': {6,7,8}}, 'AATTCC<pr>ATC</pr>GG')
        self.t({'pr': {1,2,3,6,7,8}}, 'A<pr>ATT</pr>CC<pr>ATC</pr>GG')
        self.t({'g': {1,2,3}, 'pr': {6,7,8}}, 'A<g>ATT</g>CC<pr>ATC</pr>GG')

    def test_adjacent(self):
        self.t({'pr': {1,2,3}, 'g': {4,5,6}}, 'A<pr>ATT</pr><g>CCA</g>TCGG')
        self.t({'pr': {1,2,3,7,8,9}, 'g': {4,5,6}},
                'A<pr>ATT</pr><g>CCA</g><pr>TCG</pr>G')

    def test_overlapping_different_labels(self):
        f = formatting.tag_seq_overlap({'pr': {1,2,3}, 'g': {3,4,5}},
                'AATTCCATCGG')
        self.assertIn(f, ['A<pr>ATT</pr><g>CC</g>ATCGG',
            'A<pr>AT</pr><g>TCC</g>ATCGG'])

    def test_at_ends(self):
        self.t({'pr': {0,1,2,8,9,10}}, '<pr>AAT</pr>TCCAT<pr>CGG</pr>')
        self.t({'pr': {0,1,2,8,9,10}, 'g': {4,5,6}},
                '<pr>AAT</pr>T<g>CCA</g>T<pr>CGG</pr>')

    def tearDown(self):
        # Re-enable logging
        logging.disable(logging.NOTSET)



"""Functions for preparing an alignment for input.

This includes downloading, curating, and aligning sequences.
"""

from collections import Counter
from collections import OrderedDict
import logging
import os
import random

from adapt.prepare import align
from adapt.prepare import cluster
from adapt.prepare import ncbi_neighbors
from adapt.utils import seq_io

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


def prepare_for(taxid, segment, ref_accs, out,
        aln_memoizer=None, aln_stat_memoizer=None, 
        sample_seqs=None, filter_warn=0.25, min_seq_len=150,
        min_cluster_size=2, prep_influenza=False, years_tsv=None,
        cluster_threshold=0.1, accessions_to_use=None, 
        sequences_to_use=None, meta_filt=None, meta_filt_against=None):
    """Prepare an alignment for a taxonomy.

    This does the following:
        1) Downloads a list of neighbors.
        2) Downloads sequences of those neighbors.
        3) Curates the sequences.
        4) Clusters the unaligned sequences.
        5) Aligns the sequences, with one alignment per cluster.

    Args:
        taxid: taxonomic ID from NCBI, for which to download
            sequences
        segment: only use sequences of this segment (ignored if set
            to '' or None, e.g., if taxid is unsegmented)
        ref_accs: list of accessions of reference sequences to use for curation
        out: path to direcotry in which write FASTA files of aligned
            sequences (one per cluster)
        aln_memoizer: AlignmentMemoizer to use for memoizing alignments;
            or None to not memoize
        aln_stat_memoizer: AlignmentStatMemoizer to use for memoization
            of alignment statistic values; or None to not memoize
        sample_seqs: fetch all accessions, and then randomly select
            SAMPLE_SEQS of them with replacement and design from
            these; if None (default), do not sample the input
        filter_warn: raise a warning if the fraction of sequences that
            are filtered out during curation is greater than or equal to
            this float
        min_seq_len: toss any sequences whose length is less than
            this. The value should be at least as largest as the
            number of hash values used to produce a sequence signature
            for clustering (in cluster.cluster_with_minhash_signatures());
            otherwise, that function will be unable to form a signature
        min_cluster_size: toss all sequences from any clusters with
            less than this number of sequences; in other words, ignore
            clusters less than this size
        prep_influenza: if True, assume taxid represents an Influenza
            A or B virus taxonomy, and fetch sequences using NCBI's
            Influenza database
        years_tsv: if set, a path to a TSV file to which this will write
            a year (column 2) for each sequence written to out (column 1);
            this is only done for influenza sequences (if prep_influenza
            is True)
        cluster_threshold: maximum inter-cluster distance to merge clusters, in
            average nucleotide dissimilarity (1-ANI, where ANI is
            average nucleotide identity); higher results in fewer
            clusters
        accessions_to_use: if set, a collection of accessions to use instead
            of fetching neighbors for taxid
        sequences_to_use: if set, a dict of sequences to use instead of
            fetching them for taxid; note that this does not perform
            curation on these sequences
        meta_filt: tuple of 2 dictionaries where the keys are any of 'country', 'year',
            'entry_create_year', 'taxid' and values for the first are a collection 
            of what to include or True to indicate that the metadata must exist and
            the second are what to exclude.
        meta_filt_against: tuple of 2 dictionaries where the keys are any of 'country', 
            'year', 'entry_create_year', 'taxid' and values for the first are a 
            collection of what to include in accessions to be specific against and
            the second are what to exclude.

    Returns:
        number of clusters, set of accessions filtered out by meta_filt_against
    """
    logger.info(("Preparing an alignment for tax %d (segment: %s) with "
        "references %s") % (taxid, segment, ref_accs))

    specific_against_metadata_acc = set()

    if taxid in [11320, 11520, 11552]:
        # Represents influenza
        prep_influenza = True

    if sequences_to_use is not None:
        seqs_unaligned = sequences_to_use
        seqs_unaligned_curated = seqs_unaligned
    else:
        if accessions_to_use is not None:
            neighbors = [ncbi_neighbors.Neighbor(acc, None, None, None, None,
                segment) for acc in accessions_to_use]
        else:
            # Download neighbors for taxid
            if prep_influenza:
                neighbors = ncbi_neighbors.construct_influenza_genome_neighbors(taxid)
            else:
                neighbors = ncbi_neighbors.construct_neighbors(taxid)

            # Filter neighbors by segment
            if segment != None and segment != '':
                neighbors = [n for n in neighbors if n.segment == segment]

        num_unique_acc = len(set(n.acc for n in neighbors))
        logger.info(("There are %d neighbors (%d with unique accessions)"),
                len(neighbors), num_unique_acc)

        # Filter out anything that does not have a year if years_tsv is defined
        if years_tsv is not None:
            if meta_filt:
                if 'year' not in meta_filt[0]:
                    meta_filt[0]['year'] = True
            else:
                meta_filt = ({'year': True}, None)
        
        # Fetch metadata, add it to neighbors, and filter out ones that do
        # not fit the filters
        if meta_filt is not None or meta_filt_against is not None:
            neighbors, specific_against_metadata_acc = ncbi_neighbors.add_metadata_to_neighbors_and_filter(
                neighbors, meta_filt, meta_filt_against)

        if len(neighbors) == 0:
            if segment != None and segment != '':
                raise Exception(("No sequences were found for taxid %d and "
                    "segment '%s'") % (taxid, segment))
            else:
                raise Exception(("No sequences were found for taxid %d") % taxid)

        if sample_seqs is not None:
            # Randomly sample sample_seqs from neighbors
            if sample_seqs > len(neighbors):
                logger.warning(("sample_seqs for tax %d (segment: %s) is %d and "
                    "is greater than the number of sequences available (%d); "
                    "still subsampling with replacement, but this might be "
                    "unintended") %
                    (taxid, segment, sample_seqs, len(neighbors)))
            # Sample accessions, rather than neighbors, because an accession can
            # show up multiple times in neighbors due to having multiple RefSeq
            # entries
            acc_to_sample = list(set([n.acc for n in neighbors]))
            acc_to_sample.sort()
            acc_to_fetch = random.choices(acc_to_sample, k=sample_seqs)
            neighbors = [n for n in neighbors if n.acc in acc_to_fetch]
            # Because this is sampling with replacement, an accession may
            # show up multiple times in the sampling; to keep this multiplicity
            # going forward, track its count
            acc_count = Counter(acc_to_fetch)
            logger.info(("After sampling with replacement to %d sequences, there "
                "are %d unique accessions; after curation, this will repeat ones "
                "in the input that were sampled more than once (the final number, "
                "even after adding multiplicity, may be less than %d if accessions "
                "are filtered out during curation)"),
                sample_seqs, len(set(acc_to_fetch)), sample_seqs)
        else:
            # Only fetch each accession once
            acc_to_fetch = list(set([n.acc for n in neighbors]))
            acc_count = Counter(acc_to_fetch)   # 1 for each accession

        # Fetch FASTAs for the neighbors; also do so for ref_accs if ones
        # are not included
        # Keep track of whether they were added, so that they can be removed
        # later on
        added_ref_accs_to_fetch = []
        for ref_acc in ref_accs:
            if ref_acc not in acc_to_fetch:
                acc_to_fetch += [ref_acc]
                added_ref_accs_to_fetch += [ref_acc]
        seqs_unaligned_fp = ncbi_neighbors.fetch_fastas(acc_to_fetch)
        seqs_unaligned = align.read_unaligned_seqs(seqs_unaligned_fp)
        # Delete temporary file
        os.unlink(seqs_unaligned_fp.name)

        # Toss sequences that are too short
        seqs_too_short = set()
        for name, seq in seqs_unaligned.items():
            if len(seq) < min_seq_len:
                seqs_too_short.add(name)
        if len(seqs_too_short) > 0:
            logger.warning(("Ignoring the following sequences, whose lengths "
                "are <%d nt long: %s"), min_seq_len,
                str(sorted(list(seqs_too_short))))
        for name in seqs_too_short:
            del seqs_unaligned[name]

        # Curate against ref_accs
        seqs_unaligned_curated = align.curate_against_ref(
            seqs_unaligned, ref_accs, asm=aln_stat_memoizer,
            remove_ref_accs=added_ref_accs_to_fetch)

        # An accession can show up multiple times if there was sampling with
        # replacement, but the dicts above (seqs_unaligned and
        # seqs_unaligned_curate) are key'd on accession; replicate the
        # multiplicity here but renaming an accession that should appear multiple
        # times from [accession.version] to [accession.version]-{1,2,...}
        seqs_unaligned_curated_with_multiplicity = OrderedDict()
        for accver, seq in seqs_unaligned_curated.items():
            acc = accver.split('.')[0]
            assert acc_count[acc] > 0
            if acc_count[acc] == 1:
                seqs_unaligned_curated_with_multiplicity[accver] = seq
            else:
                logger.debug(("Adding multiplicity %d to accession %s") %
                        (acc_count[acc], acc))
                for i in range(1, acc_count[acc] + 1):
                    new_name = accver + '-' + str(i)
                    seqs_unaligned_curated_with_multiplicity[new_name] = seq
        seqs_unaligned_curated = seqs_unaligned_curated_with_multiplicity

        if sample_seqs is not None:
            logger.info(("After adding multiplicity, there are %d sequences"),
                len(seqs_unaligned_curated))

        # When determining how many were filtered out, we care about acc_to_fetch
        # (which has multiplicity in case of sampling with replacement) minus any
        # removed reference sequences
        frac_filtered = 1.0 - float(len(seqs_unaligned_curated)) / \
                (len(acc_to_fetch) - len(added_ref_accs_to_fetch))
        if frac_filtered >= filter_warn:
            logger.warning(("A fraction %f of sequences were filtered out "
                "during curation for tax %d (segment: %s) using references %s") %
                (frac_filtered, taxid, segment, ref_accs))

        # Check if there are no sequences left; if that's the case, warn
        # and try just a reference sequence
        if len(seqs_unaligned_curated) == 0:
            # Find a reference genome that was downloaded
            ref_accver_used = None
            for accver, seq in seqs_unaligned.items():
                if accver.split('.')[0] in ref_accs:
                    seqs_unaligned_curated[accver] = seq
                    ref_accver_used = accver
                    break
            if len(seqs_unaligned_curated) > 0:
                logger.critical(("No sequences remained after curation, so "
                    "proceeding with design from a single reference sequence "
                    "(%s)") % (ref_accver_used))
            else:
                raise Exception("No sequences are available for design")

    # Produce clusters of unaligned sequences
    logger.info(("Clustering %d sequences"), len(seqs_unaligned_curated))
    clusters = cluster.cluster_with_minhash_signatures(
        seqs_unaligned_curated, threshold=cluster_threshold)

    # Throw away clusters (and the sequences in them) that are too small;
    # but only throw away clusters if at least 1 will remain
    # Note that clusters is sorted in descending order of cluster size
    seqs_from_small_clusters = set()
    cluster_to_remove = None
    for cluster_idx, seqs_in_cluster in enumerate(clusters):
        if len(seqs_in_cluster) < min_cluster_size:
            seqs_from_small_clusters.update(seqs_in_cluster)
            if cluster_to_remove is None:
                cluster_to_remove = cluster_idx
    if cluster_to_remove is not None and cluster_to_remove > 0:
        # There is a cluster to remove, and it is not the first (i.e.,
        # one will remain)
        logger.warning(("Removing clusters that are too small (<%d), and "
            "ignoring all the sequences they contain: %s"), min_cluster_size,
            str(sorted(list(seqs_from_small_clusters))))
        clusters = clusters[:cluster_to_remove]

    # Align the curated sequences, with one alignment per cluster
    for cluster_idx, seqs_in_cluster in enumerate(clusters):
        logger.info(("Aligning sequences in cluster %d (of %d), with %d "
            "sequences"), cluster_idx + 1, len(clusters), len(seqs_in_cluster))

        # Produce a dict of the sequences in cluster_idx
        seqs_unaligned_curated_in_cluster = OrderedDict()
        for name, seq in seqs_unaligned_curated.items():
            if name in seqs_in_cluster:
                seqs_unaligned_curated_in_cluster[name] = seq

        # Align the sequences in this cluster
        seqs_aligned = align.align(seqs_unaligned_curated_in_cluster,
            am=aln_memoizer)

        # Write a fasta file of aligned sequences
        fasta_file = os.path.join(out, str(cluster_idx) + '.fasta')
        seq_io.write_fasta(seqs_aligned, fasta_file)

    # Write the years for each sequence, if requested
    if years_tsv:
        year_for_acc = {neighbor.acc: neighbor.metadata['year'] for
                neighbor in neighbors}
        all_seq_names = set().union(*clusters)
        with open(years_tsv, 'w') as fw:
            for name in all_seq_names:
                # name is [accession].[version] (or [accession].[version]-[n]
                # in case of multiplicity); extract the accession
                acc = name.split('.')[0]
                fw.write('\t'.join([name, str(year_for_acc[acc])]) + '\n')

    return len(clusters), specific_against_metadata_acc


def fetch_sequences_for_taxonomy(taxid, segment):
    """Fetch list of sequences for a NCBI taxonomy.

    Args:
        taxid: taxonomic ID from NCBI, for which to download
            sequences
        segment: only use sequences of this segment (ignored if set
            to '' or None, e.g., if taxid is unsegmented)

    Returns:
        dict mapping sequence header to sequence string
    """
    if taxid in [11320, 11520, 11552]:
        # Represents influenza
        prep_influenza = True
    else:
        prep_influenza = False

    # Download neighbors for taxid
    if prep_influenza:
        neighbors = ncbi_neighbors.construct_influenza_genome_neighbors(taxid)
    else:
        neighbors = ncbi_neighbors.construct_neighbors(taxid)

    # Filter neighbors by segment
    if segment != None and segment != '':
        neighbors = [n for n in neighbors if n.segment == segment]

    # Only fetch each accession once
    acc_to_fetch = list(set([n.acc for n in neighbors]))

    if len(acc_to_fetch) == 0:
        logger.critical(("There are 0 accessions for tax %d (segment: %s)"),
                taxid, segment)

    seqs_unaligned = fetch_sequences_for_acc_list(acc_to_fetch)

    return seqs_unaligned


def fetch_sequences_for_acc_list(acc_to_fetch):
    """Fetch list of sequences given a list of accessions

    Args:
        acc_to_fetch: list of accessions

    Returns:
        dict mapping sequence header to sequence string
    """
    seqs_unaligned_fp = ncbi_neighbors.fetch_fastas(acc_to_fetch)
    seqs_unaligned = align.read_unaligned_seqs(seqs_unaligned_fp)
    # Delete temporary file
    os.unlink(seqs_unaligned_fp.name)

    return seqs_unaligned



"""Utilities for working with genome neighbors (complete genomes) from NCBI.
"""

from collections import defaultdict
import datetime
import gzip
import http.client
import logging
import random
import re
import socket
import tempfile
import time
import urllib.parse
import urllib.request
from xml.dom import minidom
from os import unlink

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


# Global variable for API key
ncbi_api_key = None


def urlopen_with_tries(url, initial_wait=5, rand_wait_range=(1, 60),
        max_num_tries=10, timeout=60, read=False):
    """
    Open a URL via urllib with repeated tries.

    Often calling urllib.request.urlopen() fails with HTTPError, especially
    if there are multiple processes calling it. The reason is that NCBI
    has a cap on the number of requests per unit time, and the error raised
    is 'HTTP Error 429: Too Many Requests'.

    Args:
        url: url to open
        initial_wait: number of seconds to wait in between the first two
            requests; the wait for each subsequent request doubles in time
        rand_wait_range: tuple (a, b); in addition to waiting an amount of
            time that grows exponentially (starting with initial_wait), also
            wait a random number of seconds between a and b (inclusive).
            If multiple processes are started simultaneously, this helps to
            avoid them waiting on the same cycle
        max_num_tries: maximum number of requests to attempt to make
        timeout: timeout in sec before retrying
        read: also try to read the opened URL, and return the results;
            if this raises an HTTPException, the call will be retried

    Returns:
        result of urllib.request.urlopen(); unless read is True, in which
        case it is the data returned by reading the url
    """
    num_tries = 0
    while num_tries < max_num_tries:
        try:
            num_tries += 1
            logger.debug(("Making request to open url: %s"), url)
            r = urllib.request.urlopen(url, timeout=timeout)
            if read:
                raw_data = r.read()
                return raw_data
            else:
                return r
        except (urllib.error.HTTPError, http.client.HTTPException,
                urllib.error.URLError, socket.timeout):
            if num_tries == max_num_tries:
                # This was the last allowed try
                logger.warning(("Encountered HTTPError or HTTPException or "
                    "URLError or timeout %d times (the maximum allowed) when "
                    "opening url: %s"),
                    num_tries, url)
                raise
            else:
                # Pause for a bit and retry
                wait = initial_wait * 2**(num_tries - 1)
                rand_wait = random.randint(*rand_wait_range)
                total_wait = wait + rand_wait
                logger.info(("Encountered HTTPError or HTTPException or "
                    "URLError or timeout when opening url; sleeping for %d "
                    "seconds, and then trying again"), total_wait)
                time.sleep(total_wait)
        except:
            logger.warning(("Encountered unexpected error while opening "
                "url: %s"), url)
            raise


def ncbi_neighbors_url(taxid):
    """Construct URL for downloading list of genome neighbors.

    Args:
        taxid: taxonomic ID to download neighbors for

    Returns:
        str representing download URL
    """
    params = [('taxid', taxid), ('cmd', 'download2')]
    if ncbi_api_key is not None:
        params += [('api_key', ncbi_api_key)]
    params_url = urllib.parse.urlencode(params)
    url = 'https://www.ncbi.nlm.nih.gov/genomes/GenomesGroup.cgi?%s' % params_url
    return url


def ncbi_reference_url(taxid):
    """Construct URL for downloading list of genome references.

    Args:
        taxid: taxonomic ID to download references for

    Returns:
        str representing download URL
    """
    params = [('taxid', taxid), ('cmd', 'download1')]
    if ncbi_api_key is not None:
        params += [('api_key', ncbi_api_key)]
    params_url = urllib.parse.urlencode(params)
    url = 'https://www.ncbi.nlm.nih.gov/genomes/GenomesGroup.cgi?%s' % params_url
    return url


def read_taxid_from_ncbi_url(url):
    """Read the taxid from a URL.

    Args:
        url: str representing URL for downloading NCBI neighbors

    Returns:
        taxid in url, or None if no taxid is found
    """
    url_query = urllib.parse.urlparse(url).query
    queries = urllib.parse.parse_qs(url_query)
    if 'taxid' not in queries:
        return None
    return int(queries['taxid'][0])


def fetch_neighbors_table(taxid):
    """Fetch genome neighbors table from NCBI.

    Args:
        taxid: taxonomic ID to download neighbors for

    Yields:
        lines, where each line is from the genome neighbors
        table and each line is a str
    """
    logger.debug(("Fetching table of neighbors for tax %d") % taxid)

    url = ncbi_neighbors_url(taxid)
    r = urlopen_with_tries(url)

    # For some taxids, NCBI redirects to a different taxid and we are
    # then unable to read the contents of the response (because it is
    # HTML, and not the expected text format)
    # Check if this occurred by seeing if the taxid in the url of
    # the response is different from what was requested; if it is,
    # then re-fetch the url with the new taxid (only try this once, otherwise
    # we could end up in a loop of redirects)
    taxid_in_url = read_taxid_from_ncbi_url(r.geturl())
    if taxid_in_url != taxid:
        logger.warning(("The neighbors table for taxid %d is being redirected "
            "to taxid %d"), taxid, taxid_in_url)
        new_taxid = taxid_in_url
        url = ncbi_neighbors_url(new_taxid)
        r = urlopen_with_tries(url)

    raw_data = r.read()
    for line in raw_data.decode('utf-8').split('\n'):
        line_rstrip = line.rstrip()
        if line_rstrip != '':
            yield line_rstrip


def fetch_references(taxid):
    """Fetch genome references list from NCBI.

    Args:
        taxid: taxonomic ID to download references for

    Yields:
        lines, where each line is from the genome references
        list and each line is a str
    """
    logger.debug(("Fetching list of references for tax %d") % taxid)

    url = ncbi_reference_url(taxid)
    r = urlopen_with_tries(url)

    # For some taxids, NCBI redirects to a different taxid and we are
    # then unable to read the contents of the response (because it is
    # HTML, and not the expected text format)
    # Check if this occurred by seeing if the taxid in the url of
    # the response is different from what was requested; if it is,
    # then re-fetch the url with the new taxid (only try this once, otherwise
    # we could end up in a loop of redirects)
    taxid_in_url = read_taxid_from_ncbi_url(r.geturl())
    if taxid_in_url != taxid:
        logger.warning(("The references list for taxid %d is being redirected "
            "to taxid %d"), taxid, taxid_in_url)
        new_taxid = taxid_in_url
        url = ncbi_reference_url(new_taxid)
        r = urlopen_with_tries(url)

    raw_data = r.read()
    for line in raw_data.decode('utf-8').split('\n'):
        line_rstrip = line.rstrip()
        if line_rstrip != '':
            yield line_rstrip


def ncbi_influenza_genomes_url(database='genomeset'):
    """Construct URL for downloading NCBI influenza genomes database.

    Args:
        database: db to use; 'genomeset' or 'influenza_na'

    Returns:
        str representing download URL
    """
    url = 'ftp://ftp.ncbi.nih.gov/genomes/INFLUENZA/'
    assert database in ['genomeset', 'influenza_na']
    url += database + '.dat.gz'
    return url


def fetch_influenza_genomes_table(species_name, database):
    """Fetch influenza genome table from NCBI.

    Args:
        species_name: filter to keep only lines that contain this species
            name
        database: db to use; 'genomeset' or 'influenza_na'

    Yields:
        lines, where each line is from the genome database table and
        each line is a str
    """
    logger.debug(("Fetching table of influenza genomes for species %s") %
        species_name)
    species_name_lower = species_name.lower()

    url = ncbi_influenza_genomes_url(database)
    r = urlopen_with_tries(url)
    raw_data = gzip.GzipFile(fileobj=r).read()
    for line in raw_data.decode('utf-8').split('\n'):
        line_rstrip = line.rstrip()
        if line_rstrip != '':
            if species_name_lower in line_rstrip.lower():
                yield line_rstrip


def ncbi_fasta_download_url(accessions):
    """Construct URL for downloading FASTA sequence.

    Args:
        accessions: collection of accessions to download sequences for

    Returns:
        str representing download URL
    """
    ids = ','.join(accessions)
    params = [('id', ids), ('db', 'nuccore'), ('rettype', 'fasta'),
                ('retmode', 'text')]
    if ncbi_api_key is not None:
        params += [('api_key', ncbi_api_key)]
    # Use safe=',' to not encode ',' as '%2'
    params_url = urllib.parse.urlencode(params, safe=',')
    url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?%s' % params_url
    return url


def fetch_fastas(accessions, batch_size=100, reqs_per_sec=2):
    """Download sequences for accessions.

    Entrez enforces a limit of ~3 requests per second (or else it
    will return a 'Too many requests' error); to avoid this, this
    aims for ~2 requests per second. To use up to 10 requests per second,
    request an API key from Entrez.

    Args:
        accessions: collection of accessions to download sequences for
        batch_size: number of accessions to download in each batch
        reqs_per_sec: number of requests per second to allow

    Returns:
        tempfile object containing the sequences in fasta format
    """
    logger.debug(("Fetching fasta files for %d accessions") % len(accessions))

    if ncbi_api_key is not None:
        # Using an API keys allows more requests per second (up to 10)
        reqs_per_sec = 7

    # Make temp file
    fp = tempfile.NamedTemporaryFile(delete=False)

    # Download sequences in batches
    for i in range(0, len(accessions), batch_size):
        batch = accessions[i:(i + batch_size)]
        url = ncbi_fasta_download_url(batch)
        raw_data = urlopen_with_tries(url, read=True)
        for line in raw_data.decode('utf-8').split('\n'):
            fp.write((line + '\n').encode())
        time.sleep(1.0/reqs_per_sec)

    # Closes the file so that it can be reopened on Windows
    fp.close()

    return fp


def ncbi_xml_download_url(accessions):
    """Construct URL for downloading GenBank XML.

    Args:
        accessions: collection of accessions to download XML for

    Returns:
        str representing download URL
    """
    ids = ','.join(accessions)
    params = [('id', ids), ('db', 'nuccore'), ('retmode', 'xml')]
    if ncbi_api_key is not None:
        params += [('api_key', ncbi_api_key)]
    # Use safe=',' to not encode ',' as '%2'
    params_url = urllib.parse.urlencode(params, safe=',')
    url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?%s' % params_url
    return url


def fetch_xml(accessions, batch_size=100, reqs_per_sec=2):
    """Download XML for accessions.

    Entrez enforces a limit of ~3 requests per second (or else it
    will return a 'Too many requests' error); to avoid this, this
    aims for ~2 requests per second. To use up to 10 requests per second,
    request an API key from Entrez.

    Args:
        accessions: collection of accessions to download XML for
        batch_size: number of accessions to download in each batch
        reqs_per_sec: number of requests per second to allow

    Returns:
        tempfile object containing the downloaded XML data
    """
    # Make temp file
    fp = tempfile.NamedTemporaryFile(delete=False)

    # Only write the header once; otherwise, it will be written for each
    # beach, and then the file will not be able to be parsed
    def is_xml_header(line):
        return (line.startswith('<?xml ') or line.startswith('<!DOCTYPE '))

    # Download in batches
    for i in range(0, len(accessions), batch_size):
        batch = accessions[i:(i + batch_size)]
        url = ncbi_xml_download_url(batch)
        raw_data = urlopen_with_tries(url, read=True)
        for line in raw_data.decode('utf-8').split('\n'):
            if i > 0 and is_xml_header(line):
                # Only write XML header for the first batch (i == 0)
                continue
            if i > 0 and '<GBSet>' in line:
                # Do not write GBSet open after the first batch
                line = line.replace('<GBSet>', '')
            if '</GBSet>' in line:
                # Never write GBSet close until the end
                line = line.replace('</GBSet>', '')
            fp.write((line + '\n').encode())
        time.sleep(1.0/reqs_per_sec)

    # Write the GBSet close
    fp.write(('</GBSet>' + '\n').encode())

    # Closes the file so that it can be reopened on Windows
    fp.close()

    return fp


class Neighbor:
    """Immutable representation of a genome neighbor."""

    def __init__(self, acc, refseq_acc, hosts, lineage, tax_name, segment,
            metadata={}):
        self.acc = acc
        self.refseq_acc = refseq_acc
        self.hosts = hosts
        self.lineage = lineage
        self.tax_name = tax_name
        self.segment = segment
        self.metadata = metadata

    def _list_of_attrs(self):
        """List of all attributes (except the accession)."""
        return [self.refseq_acc, self.hosts, self.lineage, self.tax_name,
            self.segment, self.metadata]

    def __eq__(self, other):
        return (self.acc == other.acc and
                self.refseq_acc == other.refseq_acc and
                sorted(self.hosts) == sorted(other.hosts) and
                self.lineage == other.lineage and
                self.tax_name == other.tax_name and
                self.segment == other.segment,
                self.metadata == other.metadata)

    def __repr__(self):
        return ';'.join('"' + str(s) + '"' for s in
            [self.acc] + self._list_of_attrs())

    def __str__(self):
        return self.acc + ' : ' + ', '.join('"' + str(s) + '"' for s in
            self._list_of_attrs())


def construct_neighbors(taxid):
    """Construct Neighbor objects for all neighbors of a taxonomic ID.

    Args:
        taxid: taxonomic ID to download neighbors for

    Returns:
        list of Neighbor objects
    """
    logger.info(("Constructing a list of neighbors for tax %d") % taxid)

    expected_col_order = ['Representative', 'Neighbor', 'Host',
        'Selected lineage', 'Taxonomy name', 'Segment name']

    neighbors = []
    encountered_header = False
    for line in fetch_neighbors_table(taxid):
        if len(line.strip()) == 0:
            continue

        ls = line.split('\t')

        if line.startswith('##'):
            # Header line
            encountered_header = True
            if line.startswith('## Columns:'):
                # Verify the columns are as expected
                col_names = [n.replace('"', '') for n in ls[1:]]
                if expected_col_order != col_names:
                    raise Exception(("The order of columns in the neighbor "
                        "list does not match the expected order"))
            # Skip the header
            continue

        if not encountered_header:
            logger.critical(("Neighbors table for tax %d did not contain "
                "the expected header; it is possible that this is not a "
                "valid taxonomy ID"), taxid)

        refseq_acc = ls[0]
        acc = ls[1]
        hosts = ls[2].split(',')
        lineage = tuple(ls[3].split(','))
        tax_name = ls[4]
        segment = ls[5].replace('segment', '').strip()

        neighbor = Neighbor(acc, refseq_acc, hosts, lineage, tax_name, segment)
        neighbors += [neighbor]

    return neighbors

def construct_references(taxid):
    """Construct reference accession number list for a taxonomic ID.

    Args:
        taxid: taxonomic ID to download references for

    Returns:
        list of strings of reference accession numbers
    """
    logger.info(("Constructing a list of references for tax %d") % taxid)

    references = []
    for line in fetch_references(taxid):
        ls = line.strip()
        if len(ls) == 0:
            continue
        references.append(ls)

    if len(references) == 0:
        raise ValueError("Taxonomic ID %d does not have any references in "
                         "NCBI. Try specifying a reference accession using "
                         "--ref-accs" %taxid)
    return references


def add_metadata_to_neighbors_and_filter(neighbors, meta_filt=None, meta_filt_against=None):
    """Fetch and add metadata to neighbors, and filter them based on metadata

    This only fetches for neighbors that do not have metadata set.

    This also filters out neighbors which do not match the filters in meta_filt or 
    match a filter in meta_filt_against.

    Args:
        neighbors: collection of Neighbor objects
        meta_filt: tuple of 2 dictionaries where the keys are any of 'country', 'year',
            'entry_create_year', 'taxid' and values for the first are a collection 
            of what to include or True to indicate that the metadata must exist and
            the second are what to exclude.
        meta_filt_against: tuple of 2 dictionaries where the keys are any of 'country', 
            'year', 'entry_create_year', 'taxid' and values for the first are a 
            collection of what to include in accessions to be specific against and
            the second are what to exclude.

    Returns:
        neighbors with metadata included (excluding the ones filtered out), and
            accession numbers of neighbors for the design to be specific against 
            (as specified by meta_filt_against)
    """
    # Fetch metadata for each neighbor without metadata
    to_fetch = set(n.acc for n in neighbors if n.metadata == {})
    if len(to_fetch) > 0:
        metadata = fetch_metadata(to_fetch)
    else:
        metadata = {}

    # Keep track of the number of neighbors filtered out by meta_filt_against and not 
    # meta_filt, as these may or may not be unintentional
    only_in_against = 0
    # Keep track of which neighbors don't match meta_filt or match meta_filt_against,
    # as these should not be included in the design
    acc_to_exclude = set()
    # Keep track of which neighbors match meta_filt_against, as the design should
    # be specific against these accessions
    specific_against_metadata_acc = set()

    # The first value of the filter is what the metadata should equal; the second is
    # what the metadata should not equal
    if meta_filt:
        meta_filt_eq, meta_filt_neq = meta_filt
    if meta_filt_against:
        meta_filt_against_eq, meta_filt_against_neq = meta_filt_against
    for neighbor in neighbors:
        # Set metadata for neighbor if it was not previously set
        if neighbor.acc in to_fetch:
            neighbor.metadata = metadata[neighbor.acc]

        # If there is a filter for metadata, exclude any neighbors that don't match it 
        # in the design
        if meta_filt:
            for key, value in meta_filt_eq.items():
                if value is True and neighbor.metadata[key] is None:
                    # Filter says that the metadata must exist, and it doesn't for this 
                    # neighbor, so exclude
                    acc_to_exclude.add(neighbor.acc)
                elif neighbor.metadata[key] not in value:
                    # Metadata doesn't match the list of what it should equal, so exclude
                    acc_to_exclude.add(neighbor.acc)
            for key, value in meta_filt_neq.items():
                if neighbor.metadata[key] in value:
                    # Metadata matches the list of what it shouldn't equal, so exclude
                    acc_to_exclude.add(neighbor.acc)

        # If there is a filter to be specific against metadata, add any neighbors that
        # match it to specific_against_metadata_acc and exclude them from the design
        if meta_filt_against:
            for key, value in meta_filt_against_eq.items():
                if neighbor.metadata[key] in value:
                    # Metadata matches the list of what should be designed against,
                    # so exclude if not already excluded & add to list to be specific against
                    if neighbor.acc not in acc_to_exclude:
                        acc_to_exclude.add(neighbor.acc)
                        only_in_against += 1
                    specific_against_metadata_acc.add(neighbor.acc)
            for key, value in meta_filt_against_neq.items():
                if neighbor.metadata[key] not in value:
                    # Metadata doesn't match the list of what shouldn't be designed against,
                    # so exclude if not already excluded & add to list to be specific against
                    if neighbor.acc not in acc_to_exclude:
                        acc_to_exclude.add(neighbor.acc)
                        only_in_against += 1
                    acc_to_exclude.add(neighbor.acc)
                    specific_against_metadata_acc.add(neighbor.acc)

    # Remove accessions that do not match the filters
    if len(acc_to_exclude) > 0:
        logger.info(("Leaving out %d accessions that do not match "
            "metadata filters, %d of which were only identified by "
            "specific_against_metadata_filter"), 
            len(acc_to_exclude), only_in_against)
    included_neighbors = [n for n in neighbors if n.acc not in acc_to_exclude]

    return included_neighbors, specific_against_metadata_acc


def construct_influenza_genome_neighbors(taxid):
    """Construct Neighbor objects for all influenza genomes

    According to the README on NCBI's influenza database FTP site:
    ```
    The influenza_na.dat and influenza_aa.dat files have an additional field in
    the last column to indicate the completeness of a sequence - "c" for
    complete sequences that include start and stop codons; "nc" for nearly
    complete sequences that are missing only start and/or stop codons; "p" for
    partial sequences.
    ```
    ```
    The genomeset.dat file contains information for sequences of viruses with
    a complete set of segments in full-length (or nearly full-length).  Those
    of the same virus are grouped together (using an internal group ID that is
    shown in the last column of the file) and separated by an empty line from
    those of other viruses.
    ```
    fetch_influenza_genomes_table() returns influenza_na.dat or
    genomeset.dat (as specified), filtered for a given species name.

    The columns are:
    ```
    GenBank accession number[tab]Host[tab]Genome segment number or protein name
    [tab]Subtype[tab]Country[tab]Year/month/date[tab]Sequence length
    [tab]Virus name[tab]Age[tab]Gender[tab]Completeness indicator (last field)
    ```
    (the last column is only in influenza_na.dat)

    This keeps only complete or near-complete sequences.

    Args:
        taxid: taxonomic ID for an influenza species; must be influenza A
            or B or C species

    Returns:
        list of Neighbor objects
    """
    logger.info(("Constructing a list of neighbors for influenza species "
                 "with tax %d") % taxid)

    influenza_species = {11320: 'Influenza A virus',
                         11520: 'Influenza B virus',
                         11552: 'Influenza C virus'}
    if taxid not in influenza_species:
        raise ValueError(("Taxid (%d) must be for either influenza A or "
                          "B or C virus species") % taxid)
    species_name = influenza_species[taxid]

    influenza_lineages = {11320: ('Orthomyxoviridae', 'Alphainfluenzavirus',
                                  'Influenza A virus'),
                          11520: ('Orthomyxoviridae', 'Betainfluenzavirus',
                                  'Influenza B virus'),
                          11552: ('Orthomyxoviridae', 'Gammainfluenzavirus',
                                  'Influenza C virus')}
    lineage = influenza_lineages[taxid]

    # Construct a pattern to match years in a date (1000--2999)
    year_p = re.compile('([1-2][0-9]{3})')

    # Determine the current year
    curr_year = int(datetime.datetime.now().year)

    # Choose a database to pull from; note that 11552 is only in
    # influenza_na. 11320 and 11520 are in influenza_na and
    # genomeset. influenza_na has more sequences, whereas
    # genomeset is about ~1/2 the size but more highly curated for
    # genomes
    if taxid == 11320 or taxid == 11520:
        database = 'genomeset'
    elif taxid == 11552:
        database = 'influenza_na'

    neighbors = []
    for line in fetch_influenza_genomes_table(species_name, database):
        if len(line.strip()) == 0:
            continue

        ls = line.split('\t')

        acc = ls[0]
        hosts = [ls[1]]
        segment = ls[2]
        subtype = ls[3]
        country = ls[4]
        date = ls[5]
        seq_len = int(ls[6])
        name = ls[7]

        if database == 'influenza_na':
            # These include partial sequences; cut them out - i.e., only
            # keep this sequence if it is complete or near-complete
            completeness = ls[-1]
            if completeness not in ['c', 'nc']:
                continue

        # Parse the year
        year_m = year_p.search(date)
        if year_m is None:
            # No year available; skip the sequence
            continue
        year = int(year_m.group(1))
        if year > curr_year:
            # This year is in the future (probably a typo); skip the sequence
            continue

        # Construct dict of metadata
        metadata = {'subtype': subtype, 'country': country, 'year': year,
                    'seq_len': seq_len}

        # Leave the refseq_acc as None because, in the influenza database,
        # sequences are not assigned a particular RefSeq
        neighbor = Neighbor(acc, None, hosts, lineage, name, segment,
            metadata=metadata)
        neighbors += [neighbor]

    return neighbors


def parse_genbank_xml_for_source_features(fn):
    """Parse GenBank XML to extract source features.

    Args:
        fn: path to XML file, as generated by GenBank

    Returns:
        dict {accession: [(qualifier name, qualifier value)]}
        Note that in some cases, the qualifier value can be None.
    """
    doc = minidom.parse(fn)
    def parse_xml_node_value(element, tag_name):
        els = element.getElementsByTagName(tag_name)
        if len(els) == 0:
            return None
        else:
            return els[0].firstChild.nodeValue

    source_features = defaultdict(list)

    seqs = doc.getElementsByTagName('GBSeq')
    for seq in seqs:
        accession = parse_xml_node_value(seq, 'GBSeq_primary-accession')

        create_date = parse_xml_node_value(seq, 'GBSeq_create-date')
        source_features[accession].append(('create_date', create_date))

        feature_table = seq.getElementsByTagName('GBSeq_feature-table')[0]
        for feature in feature_table.getElementsByTagName('GBFeature'):
            feature_key = parse_xml_node_value(feature, 'GBFeature_key')
            if feature_key == 'source':
                quals = feature.getElementsByTagName('GBFeature_quals')[0]
                for qualifier in quals.getElementsByTagName('GBQualifier'):
                    qual_name = parse_xml_node_value(qualifier, 'GBQualifier_name')
                    qual_value = parse_xml_node_value(qualifier, 'GBQualifier_value')
                    source_features[accession].append((qual_name, qual_value))

    return source_features


def fetch_metadata(accessions):
    """Fetch metadata from GenBank for accessions.

    This currently only parses out country and collection year.

    Args:
        accessions: collection of accessions to fetch for

    Returns:
        dict {accession: {'country': country, 'year': collection-year, 
            'entry_create_year': create-year, 'taxid': taxonomic id}}
    """
    accessions = list(set(accessions))
    logger.info(("Fetching metadata for %d accessions"), len(accessions))

    # Fetch XML and parse it for source features
    xml_tf = fetch_xml(accessions)
    source_features = parse_genbank_xml_for_source_features(xml_tf.name)

    # Construct a pattern to match years in a date (1000--2999)
    year_p = re.compile('([1-2][0-9]{3})')

    # Determine the current year
    curr_year = int(datetime.datetime.now().year)

    metadata = {}
    for accession, feats in source_features.items():
        year = None
        entry_create_year = None
        country = None
        taxid = None
        for (name, value) in feats:
            if name == 'collection_date' or name == 'create_date':
                # Parse the year
                year_m = year_p.search(value)
                if year_m is None:
                    # No year available
                    feat_year = None
                else:
                    feat_year = int(year_m.group(1))
                    if feat_year > curr_year:
                        # This year is in the future (probably a typo)
                        # Treat it as unavailable
                        feat_year = None
                    if year is not None and feat_year != year:
                        raise Exception(("Inconsistent year for "
                            "accession %s") % accession)
                if name == 'collection_date':
                    year = feat_year
                if name == 'create_date':
                    entry_create_year = feat_year
            if name == 'country':
                if country is not None and value != country:
                    raise Exception(("Inconsistent country for "
                        "accession %s") % accession)
                country = value
            if name == 'db_xref':
                # Find a string of digits that represents the taxonomic ID
                taxid = int(re.search(r"\d+", value)[0])
        metadata[accession] = {'country': country, 'year': year,
                'entry_create_year': entry_create_year, 'taxid': taxid}

    # Delete the tempfile
    unlink(xml_tf.name)

    return metadata




"""Tests for guide_search module.
"""

import logging
import random
import unittest

import numpy as np

from adapt import alignment
from adapt import guide_search
from adapt.utils import guide
from adapt.utils import index_compress

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestGuideSearcherMinimizeGuides(unittest.TestCase):
    """Tests methods in the GuideSearcherMinimizeGuides class.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.WARNING)

        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.a_seqs = ['ATCGAA', 'ATCGAT', 'AYCGAA', 'AYCGAT', 'AGCGAA']
        self.a_aln = alignment.Alignment.from_list_of_seqs(self.a_seqs)
        self.a_window_size = 5
        self.a = guide_search.GuideSearcherMinimizeGuides(self.a_aln, 4, 0, 1.0, (1, 1, 100))

        self.b_seqs = ['ATCGAA', 'ATC-AA']
        self.b_aln = alignment.Alignment.from_list_of_seqs(self.b_seqs)
        self.b_window_size = 5
        self.b = guide_search.GuideSearcherMinimizeGuides(self.b_aln, 4, 0, 1.0, (1, 1, 100))

        self.c_seqs = ['GTATCATCGGCCATGNAC',
                       'GTNNCATCGG-CATGNAC',
                       'GTNNGATAGGCCATGNAC',
                       'GAAAAAAAAAAAAAAAAC',
                       'GTATCATCGGCCATGNAC',
                       'GTATCAGCGGCCATGNAC']
        self.c_aln = alignment.Alignment.from_list_of_seqs(self.c_seqs)
        self.c_window_size = 14
        self.c = guide_search.GuideSearcherMinimizeGuides(self.c_aln, 5, 1, 1.0, (1, 1, 100))
        self.c_partial = guide_search.GuideSearcherMinimizeGuides(self.c_aln, 5, 1, 0.6,
            (1, 1, 100))

        self.d_seqs = ['GTATACGG',
                       'ACGTACGG',
                       'TACTACGG']
        self.d_aln = alignment.Alignment.from_list_of_seqs(self.d_seqs)
        self.d_window_size = 8
        self.d = guide_search.GuideSearcherMinimizeGuides(self.d_aln, 5, 0, 1.0, (1, 1, 100))

        self.e_seqs = ['GTAGACGG',
                       'ACGTACGG',
                       'TACTTCGG']
        self.e_aln = alignment.Alignment.from_list_of_seqs(self.e_seqs)
        self.e_window_size = 8
        self.e = guide_search.GuideSearcherMinimizeGuides(self.e_aln, 5, 1, 1.0, (1, 1, 100))

        self.f_seqs = ['GTNNACGN',
                       'ANNTACGN',
                       'TANTTNNN']
        self.f_aln = alignment.Alignment.from_list_of_seqs(self.f_seqs)
        self.f_window_size = 8
        self.f = guide_search.GuideSearcherMinimizeGuides(self.f_aln, 5, 1, 1.0, (1, 1, 100))

        self.g_seqs = ['GTATCATCGGCCATCNAC',
                       'CTATCACCTGCTACGNAC',
                       'ATAGCACCGGCCATGNAC',
                       'TTAGGACCGACCATGNAC']
        self.g_aln = alignment.Alignment.from_list_of_seqs(self.g_seqs)
        self.g_window_size = 18
        self.g = guide_search.GuideSearcherMinimizeGuides(self.g_aln, 5, 0, 1.0, (1, 1, 100))
        self.g_partial = guide_search.GuideSearcherMinimizeGuides(self.g_aln, 5, 0, 0.5,
            (1, 1, 100))

        self.h_seqs = ['GTATCAGCGGCCATCNACAA',
                       'GTANCACCTGCTACGNACTT',
                       'GTATCAATGNCCATGNACCC',
                       'GTATCATCCACNATGNACGG']
        self.h_aln = alignment.Alignment.from_list_of_seqs(self.h_seqs)
        self.h_window_size = 18
        self.h = guide_search.GuideSearcherMinimizeGuides(self.h_aln, 5, 1, 1.0, (0.5, 0, 1))

        self.i_seqs = ['GTATCAGCGGCCATCAACAA',
                       'GT-TCACCTGCTACGAACTT',
                       'GT-TCAATGCCCATGAACCC',
                       'GTATCATCCACCATGAACGG']
        self.i_aln = alignment.Alignment.from_list_of_seqs(self.i_seqs)
        self.i_window_size = 5
        self.i = guide_search.GuideSearcherMinimizeGuides(self.i_aln, 5, 1, 1.0, (0.5, 0, 1))

        # Skip guide clustering, which may not work well when the guides
        # are so short in these tests
        self.a.guide_clusterer = None
        self.b.guide_clusterer = None
        self.c.guide_clusterer = None
        self.d.guide_clusterer = None
        self.e.guide_clusterer = None
        self.f.guide_clusterer = None
        self.g.guide_clusterer = None
        self.h.guide_clusterer = None
        self.i.guide_clusterer = None

    def test_construct_guide_memoized_a(self):
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}}),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}}),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}}),
                         ('ATCG', {0,1,2,3}))
        self.assertIn(self.a._construct_guide_memoized(0, {0: {2,3,4}}),
                      [('ATCG', {2,3}), ('ACCG', {2,3}), ('AGCG', {4})])
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {4}}),
                         ('AGCG', {4}))

        def ic(idx):
            return index_compress.compress_mostly_contiguous(idx)

        key = (frozenset({(0, frozenset(ic({0,1,2,3,4})))}), None)
        self.assertIn(key, self.a._memoized_guides)
        self.assertIn(0, self.a._memoized_guides[key])

        key = (frozenset({(0, frozenset(ic({2,3,4})))}), None)
        self.assertIn(key, self.a._memoized_guides)
        self.assertIn(0, self.a._memoized_guides[key])

        key = (frozenset({(0, frozenset(ic({4})))}), None)
        self.assertIn(key, self.a._memoized_guides)
        self.assertIn(0, self.a._memoized_guides[key])

        self.assertEqual(self.a._construct_guide_memoized(2, {0: {0,1,2,3,4}}),
                         ('CGAA', {0,2,4}))
        self.assertEqual(self.a._construct_guide_memoized(2, {0: {3}}),
                         ('CGAT', {3}))

        key = (frozenset({(0, frozenset(ic({0,1,2,3,4})))}), None)
        self.assertIn(key, self.a._memoized_guides)
        self.assertIn(2, self.a._memoized_guides[key])

        key = (frozenset({(0, frozenset(ic({3})))}), None)
        self.assertIn(key, self.a._memoized_guides)
        self.assertIn(2, self.a._memoized_guides[key])

        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}}),
                         ('ATCG', {0,1,2,3}))
        self.assertIn(self.a._construct_guide_memoized(0, {0: {2,3,4}}),
                      [('ATCG', {2,3}), ('ACCG', {2,3}), ('AGCG', {4})])
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {4}}),
                         ('AGCG', {4}))
        self.assertEqual(self.a._construct_guide_memoized(2, {0: {0,1,2,3,4}}),
                         ('CGAA', {0,2,4}))
        self.assertEqual(self.a._construct_guide_memoized(2, {0: {3}}),
                         ('CGAT', {3}))

        self.a._cleanup_memoized_guides(2)
        for key in self.a._memoized_guides.keys():
            self.assertNotIn(2, self.a._memoized_guides[key])

        self.a._cleanup_memoized_guides(100)
        for key in self.a._memoized_guides.keys():
            self.assertNotIn(100, self.a._memoized_guides[key])

    def test_construct_guide_memoized_b(self):
        self.assertIsNone(self.b._construct_guide_memoized(0, {0: {1}}))
        self.assertEqual(self.b._construct_guide_memoized(0, {0: {0,1}}),
                         ('ATCG', {0}))
        
        self.assertIsNone(self.b._construct_guide_memoized(0, {0: {1}}))
        self.assertEqual(self.b._construct_guide_memoized(0, {0: {0,1}}),
                         ('ATCG', {0}))

    def test_construct_guide_memoized_a_with_needed(self):
        # Use the num_needed argument
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 5}),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 3}),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 5}),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 3}),
                         ('ATCG', {0,1,2,3}))

        self.a._cleanup_memoized_guides(0)
        for key in self.a._memoized_guides.keys():
            self.assertNotIn(0, self.a._memoized_guides[key])

        self.a._cleanup_memoized_guides(100)
        for key in self.a._memoized_guides.keys():
            self.assertNotIn(100, self.a._memoized_guides[key])

    def test_construct_guide_memoized_a_use_last(self):
        # Use the use_last argument
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 5}, use_last=False),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 3}, use_last=False),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 3}, use_last=True),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 3}, use_last=False),
                         ('ATCG', {0,1,2,3}))
        self.assertEqual(self.a._construct_guide_memoized(0, {0: {0,1,2,3,4}},
                            {0: 3}, use_last=True),
                         ('ATCG', {0,1,2,3}))

        self.a._cleanup_memoized_guides(0)
        for key in self.a._memoized_guides.keys():
            self.assertNotIn(0, self.a._memoized_guides[key])

    def test_find_optimal_guide_in_window(self):
        self.assertEqual(self.c._find_optimal_guide_in_window(
                            1, 1 + self.c_window_size,
                            {0: set([0,1,2,3,4,5])}, {0: 6}),
                         ('ATCGG', set([0,1,2,4,5]), 5, 5))

    def test_find_optimal_guide_in_window_at_end_boundary(self):
        self.assertEqual(self.d._find_optimal_guide_in_window(
                            0, 0 + self.d_window_size,
                            {0: set([0,1,2])}, {0: 3}),
                         ('TACGG', set([0,1,2]), 3, 3))
        self.assertEqual(self.e._find_optimal_guide_in_window(
                            0, 0 + self.e_window_size,
                            {0: set([0,1,2])}, {0: 3}),
                         ('TACGG', set([0,1,2]), 3, 3))

    def test_find_optimal_guide_in_window_none(self):
        self.assertEqual(self.f._find_optimal_guide_in_window(
                            0, 0 + self.f_window_size,
                            {0: set([0,1,2])}, {0: 3}),
                         (None, set(), None, 0))

    def test_find_optimal_guide_in_window_with_groups_1(self):
        g_opt = self.g._find_optimal_guide_in_window(0, 0 + self.g_window_size,
            {2017: {0, 2}, 2018: {1, 3}}, {2017: 0, 2018: 1})
        gd, gd_covered, gd_start, gd_score = g_opt

        # We only need to cover 1 sequence from the 2018 group ({1, 3});
        # check that at least one of these is covered
        self.assertTrue(1 in gd_covered or 3 in gd_covered)

        # Since we only need to cover 1 sequence in total, the score
        # should only be 1
        self.assertEqual(gd_score, 1)

    def test_find_guides_in_window(self):
        self.assertEqual(self.c._find_guides_in_window(
                            1, 1 + self.c_window_size),
                         set(['ATCGG', 'AAAAA']))
        self.assertIn(self.c_partial._find_guides_in_window(
                        1, 1 + self.c_window_size),
                      {frozenset(['ATCGG']), frozenset(['TCATC'])})

        self.assertIn(self.g._find_guides_in_window(
                        0, 0 + self.g_window_size),
                      [set(['TATCA', 'CCATG']), set(['CGGCC', 'TTAGG', 'CTATC'])])
        self.assertIn(self.g_partial._find_guides_in_window(
                        0, 0 + self.g_window_size),
                      [set(['TATCA']), set(['CCATG']), set(['CGGCC'])])

    def test_find_guides_with_missing_data(self):
        # The best guides are in regions with missing data, but the
        # alignment and thresholds on missing data are setup to avoid
        # guides in these regions
        self.assertEqual(self.h._find_guides_in_window(
                            0, 0 + self.h_window_size),
                         set(['CAACG', 'CACCC']))

    def test_find_guides_with_gap(self):
        # It should not be able to find a guide in a window where the only
        # possible guides overlap sequences with a gap
        with self.assertRaises(guide_search.CannotAchieveDesiredCoverageError):
            self.i._find_guides_in_window(1, 1 + self.i_window_size)

        # It should be able to find a guide in a window without a gap
        self.i._find_guides_in_window(10, 10 + self.i_window_size)

    def test_guide_is_suitable_fn(self):
        seqs = ['GTATCAAAAAATCGGCTACCCCCTCTAC',
                'CTACCAAAAAACCTGCTAGGGGGCGTAC',
                'ATAGCAAAAAAACGTCCTCCCCCTGTAC',
                'TTAGGAAAAAAGCGACCGGGGGGTCTAC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100))

        # The best guide is 'AAAAA'
        self.assertEqual(gs._find_guides_in_window(0, 28),
                         set(['AAAAA']))

        # Do not allow guides with 'AAA' in them
        def f(guide):
            if 'AAA' in guide:
                return False
            else:
                return True
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100),
            guide_is_suitable_fn=f)
        self.assertEqual(gs._find_guides_in_window(0, 28),
                         set(['CCCCC', 'GGGGG']))

    def test_with_groups(self):
        seqs = ['ATCAAATCGATGCCCTAGTCAGTCAACT',
                'ATCTTTACGATGCTCTGGTTAGCCATCT',
                'ATCTTATCGTTGGACTCGTAAGGCACCT',
                'ATCAGATCGCTGAGCTTGTGAGACAGCT',
                'TAGATCTAATCCCAGTATGGTACTTATC',
                'TAGAACTAATGGCAGTTTGGTCCTTGTC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100))

        # 4 guides are needed (3 for the first 4 sequences, 1 for the last
        # 2 sequences)
        self.assertEqual(len(gs._find_guides_in_window(0, 28)), 4)

        # Divide into groups, wanting to cover more of group 2018; now
        # we only need 1 guide from group 2010 and 1 from group 2018, so just
        # 2 guides are needed
        seq_groups = {2010: {0, 1, 2, 3}, 2018: {4, 5}}
        cover_frac = {2010: 0.1, 2018: 1.0}
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, cover_frac, (1, 1, 100),
            seq_groups=seq_groups)
        self.assertEqual(len(gs._find_guides_in_window(0, 28)), 2)

    def test_score_collection_of_guides_without_groups(self):
        seqs = ['ATCAAATCGATGCCCTAGTCAGTCAACT',
                'ATCTTTACGATGCTCTGGTTAGCCATCT',
                'ATCTTATCGTTGGACTCGTAAGGCACCT',
                'ATCAGATCGCTGAGCTTGTGAGACAGCT',
                'TAGATCTAATCCCAGTATGGTACTTATC',
                'TAGAACTAATGGCAGTTTGGTACTTGTC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100))

        # The function _score_collection_of_guides() will need to know
        # positions of guides (in _selected_guide_positions), so insert
        # these
        gs._selected_guide_positions = {'TCGAT': {6}, 'GGTAC': {18}}

        guides = ['TCGAT', 'GGTAC']
        # TCGAT covers 1 sequence (1/6) and GGTAC covers 2 sequences (2/6),
        # so the average is 0.25
        self.assertEqual(gs._score_collection_of_guides(guides), 0.25)

    def test_score_collection_of_guides_with_groups(self):
        seqs = ['ATCAAATCGATGCCCTAGTCAGTCAACT',
                'ATCTTTTCGATGCTCTGGTTAGCCATCT',
                'ATCTTATCGTTGGACTCGTAAGGCACCT',
                'ATCAGATCGCTGAGCTTGTGAGACAGCT',
                'TAGATCTAATCCCAGTATGGTACTTATC',
                'TAGAACTAATGGCAGTTTGGTTCTTGTC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        seq_groups = {2010: {0, 1, 2, 3}, 2018: {4, 5}}
        cover_frac = {2010: 0.1, 2018: 1.0}
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, cover_frac, (1, 1, 100),
            seq_groups=seq_groups)

        # The function _score_collection_of_guides() will need to know
        # positions of guides (in _selected_guide_positions), so insert
        # these
        gs._selected_guide_positions = {'TCGAT': {6}, 'GGTAC': {18}}

        guides = ['TCGAT', 'GGTAC']
        # 3 sequences are needed in total (1 from 2010 and 2 from 2018)
        # TCGAT covers 1 needed sequence from 2010 and 0 needed sequences
        # from 2018: so it covers 1/3 needed sequences
        # GGTAC covers 0 needed sequences from 2010 and 1 needed sequence
        # from 2018: so it covers 1/3 needed sequences
        # The average of these fractions (the score) is 1/3
        self.assertEqual(gs._score_collection_of_guides(guides), 1/3.0)

    def test_find_optimal_guide_with_gu_pairing(self):
        seqs = ['GTATTAACACTTCGGCTACCCCCTCTAC',
                'CTACCAACACACCTGCTAGGGGGCGTAC',
                'ATAGCAACACAACGTCCTCCCCCTGTAC',
                'TTAGGGGTGTGGCGACCGGGGGGTCTAC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        # Two guides are needed for coverage
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100),
            allow_gu_pairs=False)
        self.assertEqual(len(gs._find_guides_in_window(0, 28)), 2)

        # Only one guide is needed for coverage: 'AACAC'
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100),
            allow_gu_pairs=True)
        self.assertEqual(gs._find_guides_in_window(0, 28),
                         set(['AACAC']))

    def test_with_required_guides_full_coverage(self):
        seqs = ['ATCAAATCGATGCCCTAGTCAGTCAACT',
                'ATCTAATCGATGCTCTGGTTAGCCATCT',
                'ATCCAATCGCAGTACTCGTAAGGCACCT',
                'ATCAAATCGGTGAGCTTGTGAGACAGCT',
                'TAGAAATCGAACTAGTATGGTACTTATC',
                'TAGAAATCGTGGCAGTTTGGTTCTTGTC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        # First 5 listed guides are from the alignment and the
        # positions given are correct; last guide ('AAAAA') is made up
        # but should still be in the output
        required_guides = {'ATGCC': 9, 'ATGCT': 9, 'TCGAA': 6, 'ATCGT': 5,
                           'TTGTC': 23, 'AAAAA': 5}

        # Search with 0 mismatches and 100% coverage
        window_size = 11
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100),
            required_guides=required_guides)

        # Search in window starting at position 3
        guides_in_cover = gs._find_guides_in_window(
            3, 3 + window_size)
        # required_guides account for all sequences except the 3rd and 4th,
        # which can be both covered by 'AATCG'; the position of 'TTGTC' is
        # outside the window, so it should not be in the output
        self.assertEqual(guides_in_cover,
                         {'ATGCC', 'ATGCT', 'TCGAA', 'ATCGT', 'AAAAA', 'AATCG'})

        # Search in window starting at position 4; results should be the
        # same as above, but now use memoized values
        guides_in_cover = gs._find_guides_in_window(
            4, 4 + window_size)
        self.assertEqual(guides_in_cover,
                         {'ATGCC', 'ATGCT', 'TCGAA', 'ATCGT', 'AAAAA', 'AATCG'})

    def test_with_required_guides_partial_coverage(self):
        seqs = ['ATCAAATCGATGCCCTAGTCAGTCAACT',
                'ATCTAATCGATGCTCTGGTTAGCCATCT',
                'ATCCAATCGCAGTACTCGTAAGGCACCT',
                'ATCAAATCGGTGAGCTTGTGAGACAGCT',
                'TAGAAATCGAACTAGTATGGTACTTATC',
                'TAGAAATCGTGGCAGTTTGGTTCTTGTC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        # First 3 listed guides are from the alignment and the
        # positions given are correct; last guide ('AAAAA') is made up
        # but should still be in the output
        required_guides = {'ATGCC': 9, 'TCGAA': 6, 'TTGTC': 23,
                           'AAAAA': 5}

        # Search with 1 mismatch and 50% coverage
        window_size = 11
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 1, 0.5, (1, 1, 100),
            required_guides=required_guides)

        # Search in window starting at position 3
        guides_in_cover = gs._find_guides_in_window(
            3, 3 + window_size)
        # required_guides can account for 3 of the 6 sequences
        # the position of 'TTGTC' is outside the window, so it should not be
        # in the output
        self.assertEqual(guides_in_cover,
                         {'ATGCC', 'TCGAA', 'AAAAA'})

    def test_guide_overlaps_blacklisted_range(self):
        seqs = ['AAAAAAAAAAAAAAAAAAAA']
        aln = alignment.Alignment.from_list_of_seqs(seqs)

        blacklisted_ranges = {(3, 8), (14, 18)}

        gs = guide_search.GuideSearcherMinimizeGuides(aln, 3, 0, 1.0, (1, 1, 100),
            blacklisted_ranges=blacklisted_ranges)

        # For each position i, encode 1 if the guide (of length 3)
        # starting at i overlaps a blacklisted range, and 0 otherwise
        does_overlap = '01111111000011111100'
        for i in range(len(does_overlap)):
            if does_overlap[i] == '0':
                self.assertFalse(gs._guide_overlaps_blacklisted_range(i))
            elif does_overlap[i] == '1':
                self.assertTrue(gs._guide_overlaps_blacklisted_range(i))

    def test_optimal_guide_with_blacklisted_range(self):
        seqs = ['GTATCAAAAAATCGGCTACCCCCTCTAC',
                'CTACCAAAAAACCTGCTAGGGGGCGTAC',
                'ATAGCAAAAAAACGTCCTCCCCCTGTAC',
                'TTAGGAAAAAAGCGACCGGGGGGTCTAC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100))

        # The best guide is 'AAAAA'
        self.assertEqual(gs._find_guides_in_window(0, 28),
                         set(['AAAAA']))

        # Do not allow guides overlapping (5, 9)
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100),
            blacklisted_ranges={(5, 9)})
        self.assertEqual(gs._find_guides_in_window(0, 28),
                         set(['CCCCC', 'GGGGG']))

    def test_with_windows_of_varying_size(self):
        seqs = ['GTTCCAAAAAATCGGCTACCCCCTCTAC',
                'CTACCAAAAAACCTGCTAGGGGGCGTAC',
                'ATCGGAAAAAAACGTCCTCCCCCTGTAC',
                'TTAGGAAAAAAGCGACCGGGGGGTCTAC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100))

        # The best guides in [1, 8) are 'CCAAA' and 'GGAAA'
        self.assertEqual(gs._find_guides_in_window(1, 8),
                         set(['CCAAA', 'GGAAA']))

        # The best guide in [1, 15) is 'AAAAA'
        self.assertEqual(gs._find_guides_in_window(1, 15),
                         set(['AAAAA']))

    def test_obj_value(self):
        seqs = ['GTTCCAAAAAATCGGCTACCCCCTCTAC',
                'CTACCAAAAAACCTGCTAGGGGGCGTAC',
                'ATCGGAAAAAAACGTCCTCCCCCTGTAC',
                'TTAGGAAAAAAGCGACCGGGGGGTCTAC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100))

        # The best guides in [1, 8) are 'CCAAA' and 'GGAAA'
        guides = gs._find_guides_in_window(1, 8)
        self.assertEqual(gs.obj_value(guides), 2)

    def test_total_frac_bound_by_guides(self):
        seqs = ['GTTCCAAAAAATCGGCTACCCCCTCTAC',
                'CTACCAAAAAACCTGCTAGGGGGCGTAC',
                'ATCGGAAAAAAACGTCCTCCCCCTGTAC',
                'TTAGGAAAAAAGCGACCGGGGGGTCTAC']
        aln = alignment.Alignment.from_list_of_seqs(seqs)
        gs = guide_search.GuideSearcherMinimizeGuides(aln, 5, 0, 1.0, (1, 1, 100))

        # The best guides in [1, 8) are 'CCAAA' and 'GGAAA'
        guides = gs._find_guides_in_window(1, 8)

        # All sequences are bound by a guide
        self.assertEqual(gs.total_frac_bound_by_guides(guides), 1.0)

    def tearDown(self):
        # Re-enable logging
        logging.disable(logging.NOTSET)


class TestGuideSearcherMaximizeActivity(unittest.TestCase):
    """Tests methods in the GuideSearcherMaximizeActivity class.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.WARNING)

        # Set a random seed so hash functions are always the same
        random.seed(0)

    def make_gs(self, seqs, soft_guide_constraint=1, hard_guide_constraint=3,
            penalty_strength=0.1, algorithm='random-greedy',
            guide_is_suitable_fn=None):
        # Predict guides matching target to have activity 1, and
        # starting with 'A' to have activity 2 (otherwise, 0)
        class PredictorTest:
            def __init__(self):
                self.context_nt = 1
            def compute_activity(self, start_pos, pairs):
                y = []
                for target, guide in pairs:
                    target_without_context = target[self.context_nt:len(target)-self.context_nt]
                    if guide == target_without_context:
                        if guide[0] == 'A':
                            y += [2]
                        else:
                            y += [1]
                    else:
                        y += [0]
                return y
            def cleanup_memoized(self, pos):
                pass
        predictor = PredictorTest()

        aln = alignment.Alignment.from_list_of_seqs(seqs)
        gs = guide_search.GuideSearcherMaximizeActivity(aln, 4,
                soft_guide_constraint, hard_guide_constraint,
                penalty_strength, (1, 1, 100), algorithm=algorithm,
                predictor=predictor,
                guide_is_suitable_fn=guide_is_suitable_fn)
        return gs

    def test_obj_value_from_params(self):
        gs = self.make_gs(['ATCGATCG'])
        self.assertEqual(gs._obj_value_from_params(10, 2),
                10 - 0.1*1)

    def test_ground_set_with_activities_memoized(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           algorithm='random-greedy')

        o = gs._ground_set_with_activities_memoized(4)
        self.assertEqual(o.keys(), {'AATT', 'GGGG', 'CCCC'})
        np.testing.assert_equal(o['AATT'], np.array([2, 0, 0, 2]))
        np.testing.assert_equal(o['GGGG'], np.array([0, 1, 0, 0]))
        np.testing.assert_equal(o['CCCC'], np.array([0, 0, 1, 0]))

        o = gs._ground_set_with_activities_memoized(2)
        self.assertEqual(o.keys(), {'CGAA', 'GGGG', 'CCCC'})
        np.testing.assert_equal(o['CGAA'], np.array([1, 0, 0, 1]))
        np.testing.assert_equal(o['GGGG'], np.array([0, 1, 0, 0]))
        np.testing.assert_equal(o['CCCC'], np.array([0, 0, 1, 0]))

        # Try position 4 again
        o = gs._ground_set_with_activities_memoized(4)
        self.assertEqual(o.keys(), {'AATT', 'GGGG', 'CCCC'})
        np.testing.assert_equal(o['AATT'], np.array([2, 0, 0, 2]))
        np.testing.assert_equal(o['GGGG'], np.array([0, 1, 0, 0]))
        np.testing.assert_equal(o['CCCC'], np.array([0, 0, 1, 0]))

        gs._cleanup_memoized_ground_sets(4)

        # Try position 4 again
        o = gs._ground_set_with_activities_memoized(4)
        self.assertEqual(o.keys(), {'AATT', 'GGGG', 'CCCC'})
        np.testing.assert_equal(o['AATT'], np.array([2, 0, 0, 2]))
        np.testing.assert_equal(o['GGGG'], np.array([0, 1, 0, 0]))
        np.testing.assert_equal(o['CCCC'], np.array([0, 0, 1, 0]))

    def test_ground_set_with_activities_memoized_with_insufficient_guides(self):
        seqs = ['ATCGAATTCG']*100 + ['GGGGGGGGGG'] + ['ATCGAATTCG']*100
        gs = self.make_gs(seqs, algorithm='random-greedy')
        o = gs._ground_set_with_activities_memoized(4)
        self.assertEqual(o.keys(), {'AATT'})
        np.testing.assert_equal(o['AATT'], np.array([2]*100 + [0] + [2]*100))

    def test_activities_after_adding_guide(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           algorithm='random-greedy')
        o = gs._activities_after_adding_guide(
                [1, 1, 1, 2, 0],
                [1, 0, 2, 2, 1])
        np.testing.assert_equal(o, np.array([1, 1, 2, 2, 1]))

    def test_analyze_guides(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           algorithm='random-greedy')
        o = gs._analyze_guides(4, [3, 0, 1, 1])
        self.assertEqual(o.keys(), {'AATT', 'GGGG', 'CCCC'})
        self.assertEqual(o, {'AATT': 6/4.0, 'GGGG': 6/4.0, 'CCCC': 5/4.0})

    def test_analyze_guides_memoized(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           algorithm='random-greedy')

        o = gs._analyze_guides(4, [3, 0, 1, 1])
        self.assertEqual(o, {'AATT': 6/4.0, 'GGGG': 6/4.0, 'CCCC': 5/4.0})

        o = gs._analyze_guides(2, [10, 0, 1, 1])
        self.assertEqual(o, {'CGAA': 12/4.0, 'GGGG': 13/4.0, 'CCCC': 12/4.0})

        # Try position 4 again
        o = gs._analyze_guides(4, [3, 0, 1, 1])
        self.assertEqual(o.keys(), {'AATT', 'GGGG', 'CCCC'})
        self.assertEqual(o, {'AATT': 6/4.0, 'GGGG': 6/4.0, 'CCCC': 5/4.0})

        gs._cleanup_memoized_guides(4)

        # Try position 4 after cleanup
        o = gs._analyze_guides(4, [3, 0, 1, 1])
        self.assertEqual(o.keys(), {'AATT', 'GGGG', 'CCCC'})
        self.assertEqual(o, {'AATT': 6/4.0, 'GGGG': 6/4.0, 'CCCC': 5/4.0})

    def test_find_optimal_guide_in_window_greedy(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           algorithm='greedy')

        o = gs._find_optimal_guide_in_window(2, 8, {}, np.array([0, 0, 0, 0]))
        self.assertEqual(o, ('AATT', 4))

        o = gs._find_optimal_guide_in_window(2, 8, {'AATT'},
                np.array([0, 0, 0, 0]))
        self.assertNotEqual(o[0], 'AATT')

        o = gs._find_optimal_guide_in_window(2, 8, {'AAAA'},
                np.array([10, 1, 0, 3]))
        self.assertIn(o, [('CCCC', 2), ('CCCC', 3), ('CCCC', 4), ('CCCC', 5)])

        o = gs._find_optimal_guide_in_window(2, 8, {'ATCG'},
                np.array([2, 1, 0, 2]))
        self.assertIn(o, [('CCCC', 2), ('CCCC', 3), ('CCCC', 4), ('CCCC', 5)])

    def test_find_optimal_guide_in_window_greedy_with_high_penalty(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           penalty_strength=1,
                           algorithm='greedy')

        o = gs._find_optimal_guide_in_window(2, 8, {}, np.array([0, 0, 0, 0]))
        self.assertEqual(o, ('AATT', 4))

        # Adding any additional guide will create a negative marginal
        # contribution
        with self.assertRaises(guide_search.CannotFindPositiveMarginalContributionError):
            gs._find_optimal_guide_in_window(2, 8, {'AAAA'},
                np.array([10, 1, 0, 1]))

        # With a higher soft constraint, another guide is ok
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           soft_guide_constraint=2,
                           penalty_strength=1,
                           algorithm='greedy')
        o = gs._find_optimal_guide_in_window(2, 8, {'AAAA'},
                np.array([10, 1, 0, 1]))
        self.assertEqual(o[0], 'CCCC')

    def test_find_optimal_guide_in_window_random_greedy(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=1,
                           algorithm='random-greedy')
        o = gs._find_optimal_guide_in_window(2, 8, {}, np.array([0, 0, 0, 0]))
        self.assertEqual(o, ('AATT', 4))

        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm='random-greedy')
        o = gs._find_optimal_guide_in_window(2, 8, {'AATT'},
                np.array([0, 1, 1, 0]))
        self.assertIn(o[0], ['CGAA', 'GAAT'])

        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm='random-greedy')
        o = gs._find_optimal_guide_in_window(2, 8, {'ATCG'},
                np.array([4, 0, 0, 2]))
        self.assertIn(o[0], ['GGGG', 'CCCC'])

        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm='random-greedy')
        o = gs._find_optimal_guide_in_window(2, 8, {'GGGG'},
                np.array([0, 1, 1, 1]))
        self.assertIn(o[0], ['CGAA', 'GAAT', 'AATT'])
        o = gs._find_optimal_guide_in_window(2, 8, {'GGGG'},
                np.array([3, 3, 3, 3]))
        self.assertIsNone(o)

    def test_find_optimal_guide_in_window_random_greedy_with_high_penalty(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           penalty_strength=0.5,
                           algorithm='random-greedy')

        # Adding any additional guide will create a negative marginal
        # contribution
        o = gs._find_optimal_guide_in_window(2, 8, {'AAAA'},
                np.array([10, 0, 0, 0]))
        self.assertIsNone(o)

    def find_guides_in_window_different_hard_constraints(self, algo):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=3,
                           algorithm=algo)
        o = gs._find_guides_in_window(2, 8)
        self.assertEqual(o, {'AATT', 'GGGG', 'CCCC'})

        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm=algo)
        o = gs._find_guides_in_window(2, 8)
        self.assertIn('AATT', o)

        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=1,
                           algorithm=algo)
        o = gs._find_guides_in_window(2, 8)
        self.assertEqual(o, {'AATT'})

    def find_guides_in_window_high_penalty(self, algo):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=3,
                           penalty_strength=0.5,
                           algorithm=algo)
        o = gs._find_guides_in_window(2, 8)
        self.assertEqual(o, {'AATT'})

    def test_find_guides_in_window_different_hard_constraints_greedy(self):
        self.find_guides_in_window_different_hard_constraints('greedy')

    def test_find_guides_in_window_different_hard_constraints_random_greedy(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm='random-greedy')
        o = gs._find_guides_in_window(2, 8)
        self.assertEqual(len(o), 2) # many possibilities for o

        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=1,
                           algorithm='random-greedy')
        o = gs._find_guides_in_window(2, 8)
        self.assertEqual(o, {'AATT'})

    def test_find_guides_in_window_high_penalty_greedy(self):
        self.find_guides_in_window_high_penalty('greedy')

    def test_find_guides_in_window_high_penalty_random_greedy(self):
        self.find_guides_in_window_high_penalty('random-greedy')

    def find_guides_in_window_with_guide_is_suitable_fn(self, algo):
        def f(guide):
            if 'AA' in guide:
                return False
            else:
                return True
        
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=1,
                           algorithm=algo,
                           guide_is_suitable_fn=f)
        o = gs._find_guides_in_window(2, 8)
        self.assertEqual(len(o), 1)
        self.assertIn(list(o)[0], {'CGAA', 'GAAT', 'GGGG', 'CCCC'})

    def test_find_guides_in_window_with_guide_is_suitable_fn_greedy(self):
        self.find_guides_in_window_with_guide_is_suitable_fn('greedy')

    def test_find_guides_in_window_with_guide_is_suitable_fn_random_greedy(self):
        self.find_guides_in_window_with_guide_is_suitable_fn('random-greedy')

    def find_guides_in_window_with_gaps_and_missing_data(self, algo):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'ATCGA-TTCG',
                           'GGGGGNNNGG',
                           'AACGAATTCG'],
                           hard_guide_constraint=1,
                           algorithm=algo)
        o = gs._find_guides_in_window(2, 8)
        self.assertEqual(o, {'AATT'})

        gs = self.make_gs(['ATCGAATTCG',
                           'GGGGGGGGGG',
                           'CCCCCCCCCC',
                           'ATCGA-TTCG',
                           'GGGGGNNNGG',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm=algo)
        o = gs._find_guides_in_window(2, 8)
        self.assertEqual(len(o), 2)
        if algo == 'greedy':
            self.assertIn('AATT', o)

    def test_find_guides_in_window_with_gaps_and_missing_data_greedy(self):
        self.find_guides_in_window_with_gaps_and_missing_data('greedy')

    def test_find_guides_in_window_with_gaps_and_missing_data_random_greedy(self):
        self.find_guides_in_window_with_gaps_and_missing_data('random-greedy')

    def test_find_guides_in_window_at_endpoint(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'AGGGGGGGGA',
                           'ACCCCCCCCA',
                           'AACGAATTCG'],
                           hard_guide_constraint=1)

        # No guides should start at position 0, since it does not
        # permit a context_nt of 1 for evaluating activity
        o = gs._find_guides_in_window(0, 5)
        self.assertEqual(len(o), 1)
        self.assertIn(list(o)[0], {'TCGA', 'GGGG', 'CCCC', 'ACGA'})
        
        # No guides should end at the end of the alignment, since it
        # does not permit a context_nt of 1 for evaluating activity
        o = gs._find_guides_in_window(5, 10)
        self.assertEqual(len(o), 1)
        self.assertIn(list(o)[0], {'ATTC', 'GGGG', 'CCCC'})

        # The ranges below should not work because the test predictor
        # uses a context_nt of 1, which these do not permit
        with self.assertRaises(guide_search.CannotFindAnyGuidesError):
            gs._find_guides_in_window(0, 4)
        with self.assertRaises(guide_search.CannotFindAnyGuidesError):
            gs._find_guides_in_window(6, 10)

    def test_guide_set_activities(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGAGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm='greedy')
        # Guides should be 'AATT' and 'AGGG'
        guides = gs._find_guides_in_window(2, 8)

        activities = gs.guide_set_activities(2, 8, guides)
        np.testing.assert_equal(activities, np.array([2, 2, 0, 2]))

        activities_percentile = gs.guide_set_activities_percentile(2, 8,
                guides, [5, 50])
        self.assertEqual(activities_percentile, [0, 2])

    def test_guide_activities_expected_value(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGAGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm='greedy')
        # Guides should be 'AATT' and 'AGGG'
        guides = gs._find_guides_in_window(2, 8)

        aatt = gs.guide_activities_expected_value(2, 8, 'AATT')
        self.assertEqual(aatt, 1.0)

        aggg = gs.guide_activities_expected_value(2, 8, 'AGGG')
        self.assertEqual(aggg, 0.5)

    def test_obj_value(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGAGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           soft_guide_constraint=1,
                           penalty_strength=0.1,
                           algorithm='greedy')
        # Guides should be 'AATT' and 'AGGG'
        guides = gs._find_guides_in_window(2, 8)

        expected_activity = (2 + 2 + 0 + 2) / 4.0
        true_obj_value = expected_activity - 0.1*1
        self.assertEqual(gs.obj_value(2, 8, guides), true_obj_value)

    def test_total_frac_bound_by_guides(self):
        gs = self.make_gs(['ATCGAATTCG',
                           'GGGAGGGGGG',
                           'CCCCCCCCCC',
                           'AACGAATTCG'],
                           hard_guide_constraint=2,
                           algorithm='greedy')
        # Guides should be 'AATT' and 'AGGG'
        guides = gs._find_guides_in_window(2, 8)

        # 3 of the 4 sequences should be bound
        self.assertEqual(gs.total_frac_bound_by_guides(2, 8, guides), 0.75)

    def tearDown(self):
        # Re-enable logging
        logging.disable(logging.NOTSET)
